/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.config;

import java.io.FileReader;
import java.io.IOException;
import java.io.Serializable;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Objects;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dvirtual.database.data.ApplicationData;
import org.dussan.vaadin.dvirtual.database.data.ApplicationsGroupData;
import org.dussan.vaadin.dvirtual.database.data.UserApplicationData;
import org.dussan.vaadin.dvirtual.database.data.UserApplicationsGroupData;
import org.dussan.vaadin.dvirtual.database.data.UserData;
import org.dussan.vaadin.dvirtual.utils.LoggerUtil;
import org.slf4j.Logger;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

public class UserConfig
    implements Serializable
{

    private static final long serialVersionUID = 854891207425678489L;

    private static final Logger LOGGER = LoggerUtil.getLogger();

    private static final String DATA_VERSION = "version";

    private static final String DATA_VERSION_1 = "1";

    private static final String ERROR_CANNOT_LOAD = "Cannot load user configuration, cause: {}";

    private static final String SAX_FEATURES_EXTERNAL_GENERAL_ENTITIES = "http://xml.org/sax/features/external-general-entities";

    private static final String SPLITTER_COMMAS = ",";

    private static final String SPLITTER_SPACE = " ";

    private static final String USER = "user";

    private static final String USER_ADMINISTRATOR = "administrator";

    private static final String USER_APPLICATION_ID = "application_id";

    private static final String USER_APPLICATION_IDS = "application_ids";

    private static final String USER_APPLICATION_NAME = "application_name";

    private static final String USER_APPLICATION_NAMES = "application_names";

    private static final String USER_APPLICATIONS_GROUP_ID = "applications_group_id";

    private static final String USER_APPLICATIONS_GROUP_IDS = "applications_group_ids";

    private static final String USER_APPLICATIONS_GROUP_NAME = "applications_group_name";

    private static final String USER_APPLICATIONS_GROUP_NAMES = "applications_group_names";

    private static final String USER_DOWNLOAD = "download";

    private static final String USER_ENABLED = "enabled";

    private static final String USER_FULL_NAME = "full_name";

    private static final String USER_FULL_SCREEN = "full_screen";

    private static final String USER_ID = "id";

    private static final String USER_MAX_APPLICATIONS = "max_applications";

    private static final String USERNAME = "username";

    private static final String USER_SCREENSHOT = "screenshot";

    private static final String USER_UPLOAD = "upload";

    private transient List<UserApplicationData> userApplicationDatas = null;

    private transient List<UserApplicationsGroupData> userApplicationsGroupDatas = null;

    private transient List<UserData> userDatas = null;

    private UserConfig()
    {
        userDatas = new ArrayList<>();
        userApplicationDatas = new ArrayList<>();
        userApplicationsGroupDatas = new ArrayList<>();
    }

    public UserConfig( String configFolder, String configFile )
    {
        this();
        try (FileReader fileReader = new FileReader( Paths.get( configFolder, configFile )
                                                          .toFile() ))
        {
            DocumentBuilderFactory documentBuilderFactory = DocumentBuilderFactory.newDefaultInstance();
            documentBuilderFactory.setFeature( SAX_FEATURES_EXTERNAL_GENERAL_ENTITIES, false );
            documentBuilderFactory.setValidating( false );
            DocumentBuilder documentBuilder = documentBuilderFactory.newDocumentBuilder();
            Document document = documentBuilder.parse( new InputSource( fileReader ) );
            Element rootNode = document.getDocumentElement();

            // get data version
            String dataVersion = null;
            for ( int index = 0; index < rootNode.getAttributes()
                                                 .getLength(); index++ )
            {
                Node node = rootNode.getAttributes()
                                    .item( index );
                if ( Objects.nonNull( node ) && StringUtils.equals( node.getNodeName(), DATA_VERSION ) && StringUtils.isNotBlank( node.getTextContent() ) )
                {
                    dataVersion = StringUtils.stripToEmpty( node.getTextContent() );
                }
            }

            // user data
            if ( StringUtils.equals( dataVersion, DATA_VERSION_1 ) )
            {
                NodeList nodeList = rootNode.getElementsByTagName( USER );
                for ( int index = 0; index < nodeList.getLength(); index++ )
                {
                    Node userGroupNode = nodeList.item( index );
                    if ( userGroupNode.getNodeType() == Node.ELEMENT_NODE && userGroupNode.hasChildNodes() )
                    {
                        UserData userData = new UserData();
                        userData.setId( UserData.NEW_USER );

                        NodeList userNodeList = userGroupNode.getChildNodes();
                        for ( int i = 0; i < userNodeList.getLength(); i++ )
                        {
                            Node userChildNode = userNodeList.item( i );
                            String userNodeText = StringUtils.stripToEmpty( userChildNode.getTextContent() );
                            switch ( userChildNode.getNodeName() )
                            {
                                case USER_ADMINISTRATOR:
                                    userData.setAdministrator( Boolean.parseBoolean( userNodeText ) );
                                    break;
                                case USER_APPLICATION_ID:
                                    if ( StringUtils.isNumeric( userNodeText ) )
                                    {
                                        ApplicationData applicationData = new ApplicationData();
                                        applicationData.setId( Math.abs( Long.parseLong( userNodeText ) ) );
                                        userApplicationDatas.add( new UserApplicationData( UserApplicationData.NEW_USER_APPLICATION, userData, applicationData ) );
                                    }
                                    break;
                                case USER_APPLICATION_IDS:
                                    if ( StringUtils.isNumericSpace( StringUtils.replace( userNodeText, SPLITTER_COMMAS, SPLITTER_SPACE ) ) )
                                    {
                                        Arrays.asList( StringUtils.split( StringUtils.replace( userNodeText, SPLITTER_SPACE, SPLITTER_COMMAS ), SPLITTER_COMMAS ) )
                                              .stream()
                                              .filter( data -> StringUtils.isNotEmpty( StringUtils.stripToEmpty( data ) ) )
                                              .forEach( data -> {
                                                  ApplicationData applicationData = new ApplicationData();
                                                  applicationData.setId( Math.abs( Long.parseLong( StringUtils.strip( data ) ) ) );
                                                  userApplicationDatas.add( new UserApplicationData( UserApplicationData.NEW_USER_APPLICATION, userData, applicationData ) );
                                              } );
                                    }
                                    break;
                                case USER_APPLICATION_NAME:
                                    if ( StringUtils.isNotBlank( userNodeText ) )
                                    {
                                        ApplicationData applicationData = new ApplicationData();
                                        applicationData.setName( userNodeText );
                                        userApplicationDatas.add( new UserApplicationData( UserApplicationData.NEW_USER_APPLICATION, userData, applicationData ) );
                                    }
                                    break;
                                case USER_APPLICATION_NAMES:
                                    if ( StringUtils.isNotBlank( userNodeText ) )
                                    {
                                        Arrays.asList( StringUtils.split( userNodeText, SPLITTER_COMMAS ) )
                                              .stream()
                                              .filter( data -> StringUtils.isNotEmpty( StringUtils.stripToEmpty( data ) ) )
                                              .forEach( data -> {
                                                  ApplicationData applicationData = new ApplicationData();
                                                  applicationData.setName( StringUtils.strip( data ) );
                                                  userApplicationDatas.add( new UserApplicationData( UserApplicationData.NEW_USER_APPLICATION, userData, applicationData ) );
                                              } );
                                    }
                                    break;
                                case USER_APPLICATIONS_GROUP_ID:
                                    if ( StringUtils.isNumeric( userNodeText ) )
                                    {
                                        ApplicationsGroupData applicationsGroupData = new ApplicationsGroupData();
                                        applicationsGroupData.setId( Math.abs( Long.parseLong( userNodeText ) ) );
                                        userApplicationsGroupDatas.add( new UserApplicationsGroupData( UserApplicationsGroupData.NEW_USER_APPLICATIONS_GROUP, userData, applicationsGroupData ) );
                                    }
                                    break;
                                case USER_APPLICATIONS_GROUP_IDS:
                                    if ( StringUtils.isNumericSpace( StringUtils.replace( userNodeText, SPLITTER_COMMAS, SPLITTER_SPACE ) ) )
                                    {
                                        Arrays.asList( StringUtils.split( StringUtils.replace( userNodeText, SPLITTER_SPACE, SPLITTER_COMMAS ), SPLITTER_COMMAS ) )
                                              .stream()
                                              .filter( data -> StringUtils.isNotEmpty( StringUtils.stripToEmpty( data ) ) )
                                              .forEach( data -> {
                                                  ApplicationsGroupData applicationsGroupData = new ApplicationsGroupData();
                                                  applicationsGroupData.setId( Math.abs( Long.parseLong( StringUtils.strip( data ) ) ) );
                                                  userApplicationsGroupDatas.add( new UserApplicationsGroupData( UserApplicationsGroupData.NEW_USER_APPLICATIONS_GROUP, userData,
                                                                                                                 applicationsGroupData ) );
                                              } );
                                    }
                                    break;
                                case USER_APPLICATIONS_GROUP_NAME:
                                    if ( StringUtils.isNotBlank( userNodeText ) )
                                    {
                                        ApplicationsGroupData applicationsGroupData = new ApplicationsGroupData();
                                        applicationsGroupData.setName( userNodeText );
                                        userApplicationsGroupDatas.add( new UserApplicationsGroupData( UserApplicationsGroupData.NEW_USER_APPLICATIONS_GROUP, userData, applicationsGroupData ) );
                                    }
                                    break;
                                case USER_APPLICATIONS_GROUP_NAMES:
                                    if ( StringUtils.isNotBlank( userNodeText ) )
                                    {
                                        Arrays.asList( StringUtils.split( userNodeText, SPLITTER_COMMAS ) )
                                              .stream()
                                              .filter( data -> StringUtils.isNotEmpty( StringUtils.stripToEmpty( data ) ) )
                                              .forEach( data -> {
                                                  ApplicationsGroupData applicationsGroupData = new ApplicationsGroupData();
                                                  applicationsGroupData.setName( StringUtils.strip( data ) );
                                                  userApplicationsGroupDatas.add( new UserApplicationsGroupData( UserApplicationsGroupData.NEW_USER_APPLICATIONS_GROUP, userData,
                                                                                                                 applicationsGroupData ) );
                                              } );
                                    }
                                    break;
                                case USER_DOWNLOAD:
                                    userData.setDownload( Boolean.parseBoolean( userNodeText ) );
                                    break;
                                case USER_ENABLED:
                                    userData.setEnabled( Boolean.parseBoolean( userNodeText ) );
                                    break;
                                case USER_FULL_NAME:
                                    userData.setFullName( userNodeText );
                                    break;
                                case USER_FULL_SCREEN:
                                    userData.setFullScreen( Boolean.parseBoolean( userNodeText ) );
                                    break;
                                case USER_ID:
                                    if ( StringUtils.isNumeric( userNodeText ) )
                                    {
                                        userData.setId( Math.abs( Long.parseLong( userNodeText ) ) );
                                    }
                                    break;
                                case USER_MAX_APPLICATIONS:
                                    if ( StringUtils.isNumeric( userNodeText ) )
                                    {
                                        userData.setMaxApplications( Math.abs( Integer.parseInt( userNodeText ) ) );
                                    }
                                    break;
                                case USERNAME:
                                    userData.setUserName( userNodeText );
                                    break;
                                case USER_SCREENSHOT:
                                    userData.setScreenshot( Boolean.parseBoolean( userNodeText ) );
                                    break;
                                case USER_UPLOAD:
                                    userData.setUpload( Boolean.parseBoolean( userNodeText ) );
                                    break;
                                default:
                                    // nothing to do
                                    break;
                            }
                        }

                        // add new user data
                        if ( StringUtils.isNotBlank( userData.getUserName() ) )
                        {
                            userDatas.add( userData );
                        }
                    }
                }
            }
        }
        catch ( IOException | SAXException | ParserConfigurationException exception )
        {
            LOGGER.info( ERROR_CANNOT_LOAD, exception.getMessage() );
        }
    }

    public List<UserApplicationData> getUserApplicationDatas()
    {
        return userApplicationDatas;
    }

    public List<UserApplicationsGroupData> getUserApplicationsGroupDatas()
    {
        return userApplicationsGroupDatas;
    }

    public List<UserData> getUserDatas()
    {
        return userDatas;
    }

}
