/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.database.filters;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import javax.persistence.criteria.CriteriaBuilder;
import javax.persistence.criteria.CriteriaQuery;
import javax.persistence.criteria.Predicate;
import javax.persistence.criteria.Root;

import org.apache.commons.collections4.MapUtils;
import org.dussan.vaadin.dvirtual.database.columns.UserSessionColumns;
import org.dussan.vaadin.dvirtual.database.data.ApplicationData_;
import org.dussan.vaadin.dvirtual.database.data.ServerData_;
import org.dussan.vaadin.dvirtual.database.data.UserData_;
import org.dussan.vaadin.dvirtual.database.data.UserSessionData;
import org.dussan.vaadin.dvirtual.database.data.UserSessionData_;

public class UserSessionFilter
    implements Serializable
{

    private static final long serialVersionUID = 1679099720435634373L;

    private static final String FILTER = "%%%s%%";

    public static final <T> CriteriaQuery<T> addFilter( CriteriaBuilder criteriaBuilder, CriteriaQuery<T> criteriaQuery, Root<UserSessionData> userSession, Map<UserSessionColumns, String> filterMap )
    {
        if ( MapUtils.isNotEmpty( filterMap ) )
        {
            List<Predicate> predicates = new ArrayList<>();
            for ( Map.Entry<UserSessionColumns, String> filter : filterMap.entrySet() )
            {
                switch ( filter.getKey() )
                {
                    case ACTIVE:
                        Boolean active = Boolean.parseBoolean( filter.getValue() );
                        if ( active )
                        {
                            predicates.add( criteriaBuilder.isTrue( userSession.get( UserSessionData_.locked ) ) );
                        }
                        else
                        {
                            predicates.add( criteriaBuilder.isFalse( userSession.get( UserSessionData_.locked ) ) );
                        }
                        break;
                    case APPLICATION:
                        predicates.add( criteriaBuilder.like( criteriaBuilder.lower( userSession.get( UserSessionData_.applicationData )
                                                                                                .get( ApplicationData_.name ) ),
                                                              String.format( FILTER, filter.getValue()
                                                                                           .toLowerCase() ) ) );
                        break;
                    case NAME_FULL:
                        predicates.add( criteriaBuilder.like( criteriaBuilder.lower( userSession.get( UserSessionData_.userData )
                                                                                                .get( UserData_.fullName ) ),
                                                              String.format( FILTER, filter.getValue()
                                                                                           .toLowerCase() ) ) );
                        break;
                    case SERVER:
                        predicates.add( criteriaBuilder.like( criteriaBuilder.lower( userSession.get( UserSessionData_.serverData )
                                                                                                .get( ServerData_.name ) ),
                                                              String.format( FILTER, filter.getValue()
                                                                                           .toLowerCase() ) ) );
                        break;
                    default:
                        // nothing to do
                        break;
                }
            }
            return criteriaQuery.where( criteriaBuilder.and( predicates.toArray( Predicate[]::new ) ) );
        }
        return criteriaQuery;
    }

}
