/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.runners;

import java.util.Objects;
import java.util.concurrent.TimeUnit;

import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dvirtual.client.VirtualClient.VirtualClientAction;
import org.dussan.vaadin.dvirtual.database.Database;
import org.dussan.vaadin.dvirtual.database.data.ServerData;
import org.dussan.vaadin.dvirtual.database.data.UserSessionData;
import org.dussan.vaadin.dvirtual.i18n.Translation;
import org.dussan.vaadin.dvirtual.utils.VirtualUtil;
import org.dussan.vaadin.dvirtual.utils.LoggerUtil;
import org.dussan.vaadin.dvirtual.utils.dialog.DialogConnectingUtil;
import org.dussan.vaadin.dvirtual.xpra.client.Client;
import org.dussan.vaadin.dvirtual.xpra.connector.Connector;
import org.dussan.vaadin.dvirtual.xpra.packets.client.ClientFocusPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.client.ClientPingPacket;
import org.slf4j.Logger;

import com.vaadin.flow.component.html.Div;

public abstract class VirtualClientRunner
    implements Runnable
{

    private static final Logger LOGGER = LoggerUtil.getLogger();

    public static final long MAX_WAIT_CHILD_WINDOW_CREATED_INTERVAL = 7000; // miliseconds

    public static final long MAX_WAIT_WINDOW_CREATED_FIRST_INTERVAL = 20000; // miliseconds

    private static final int ONLY_ONE_WINDOW = 1;

    private static final int UNDEFINED_WINDOW = -1;

    private static final long MAX_CLOSE_WIDOW_WAIT_INTERVAL = 1000; // miliseconds

    private static final long MAX_WAIT_INTERVAL = 60000; // miliseconds

    private static final long MAX_WAIT_WINDOW_CREATED_SECOND_INTERVAL = 40000; // miliseconds

    private static final long NO_SLEEP_STEP_INTERVAL = 0; // miliseconds

    private static final long PROGRESS_SHOW_INTERVAL = 5000; // miliseconds

    private static final long SLEEP_PING_INTERVAL = 5000; // milisecond

    private static final long SLEEP_STEP_INTERVAL = 500; // miliseconds

    private static final String CANNOT_STOP_SERVER = "Cannot stop the server: {}:{}";

    private static final String CONNECTION_CLOSED = "Connection to the server is closed.";

    private static final String CONNECTION_OPEN = "Opening a connection to the server...";

    private long closeWaitInterval = 0;

    private long waitInterval = 0;

    private Connector connector = null;

    private Div canvasDiv = null;

    private Database database = null;

    private VirtualUtil virtualUtil = null;

    private UserSessionData userSessionData = null;

    protected VirtualClientRunner( Connector connector, Div canvasDiv )
    {
        this.connector = connector;
        this.canvasDiv = canvasDiv;
    }

    public abstract void windowCreated( boolean windowCreated );

    public void resetWaitIntervals()
    {
        closeWaitInterval = 0;
        waitInterval = 0;
    }

    public void setServerData( Database database, UserSessionData userSessionData )
    {
        this.database = database;
        this.userSessionData = userSessionData;
    }

    public VirtualUtil getVirtualUtil()
    {
        return virtualUtil;
    }

    public void setVirtualUtil( VirtualUtil virtualUtil )
    {
        this.virtualUtil = virtualUtil;
    }

    @Override
    public void run()
    {
        LOGGER.info( CONNECTION_OPEN );
        virtualUtil.setServerConnected( true );
        virtualUtil.setConnectionToServerActive( true );

        if ( Objects.nonNull( connector ) && Objects.nonNull( connector.getClient() ) && Objects.nonNull( connector.getDialogConnectingUtil() ) )
        {
            DialogConnectingUtil dialogConnectingUtil = connector.getDialogConnectingUtil();

            try
            {
                // open info dialog about connecting
                canvasDiv.getUI()
                         .ifPresent( ui -> ui.access( () -> {
                             dialogConnectingUtil.open();
                             ui.push();
                         } ) );

                // connect
                if ( connector.connect() && Objects.isNull( connector.getErrorMessage() ) )
                {
                    long pingTimeCounter = 0;
                    while ( connector.isRunning() )
                    {
                        closeWaitInterval += virtualUtil.isDrawingStarted() ? SLEEP_STEP_INTERVAL : NO_SLEEP_STEP_INTERVAL;
                        waitInterval += SLEEP_STEP_INTERVAL;
                        Client client = connector.getClient();
                        TimeUnit.MILLISECONDS.sleep( SLEEP_STEP_INTERVAL );

                        if ( virtualUtil.isNotRunningApplicationInClosingProcess() && Objects.nonNull( client.getSender() ) )
                        {
                            // send ping to server
                            if ( ( pingTimeCounter % SLEEP_PING_INTERVAL ) == 0 )
                            {
                                client.getSender()
                                      .onSend( new ClientPingPacket() );
                            }
                            pingTimeCounter += SLEEP_STEP_INTERVAL;

                            // check if window is created, if not there is an error
                            if ( closeWaitInterval >= MAX_CLOSE_WIDOW_WAIT_INTERVAL && client.isWindowCreated() && virtualUtil.isDrawingStarted() && dialogConnectingUtil.isOpened() )
                            {
                                canvasDiv.getUI()
                                         .ifPresent( ui -> ui.access( () -> {
                                             if ( virtualUtil.isNotApplicationTesting() )
                                             {
                                                 // application is running normal
                                                 if ( StringUtils.equals( dialogConnectingUtil.getMessage(), virtualUtil.getTranslation( Translation.APPLICATION_START ) )
                                                     || StringUtils.equals( dialogConnectingUtil.getMessage(), virtualUtil.getTranslation( Translation.APPLICATION_CONNECT ) ) )
                                                 {
                                                     windowCreated( true );
                                                 }
                                             }
                                             else
                                             {
                                                 // application is testing
                                                 if ( StringUtils.equals( dialogConnectingUtil.getMessage(), virtualUtil.getTranslation( Translation.APPLICATION_START ) )
                                                     || StringUtils.equals( dialogConnectingUtil.getMessage(), virtualUtil.getTranslation( Translation.APPLICATION_CONNECT ) )
                                                     || StringUtils.equals( dialogConnectingUtil.getMessage(), virtualUtil.getTranslation( Translation.APPLICATION_TESTING ) ) )
                                                 {
                                                     dialogConnectingUtil.close();
                                                 }
                                                 else
                                                 {
                                                     windowCreated( true );
                                                 }
                                             }
                                             ui.push();
                                         } ) );
                            }
                            else if ( waitInterval >= MAX_WAIT_CHILD_WINDOW_CREATED_INTERVAL && waitInterval < MAX_WAIT_INTERVAL && virtualUtil.haveToStartChildApplication()
                                && virtualUtil.canChildApplicationStarted() && virtualUtil.isNotRecreateNewWindowEnabled() )
                            {
                                virtualUtil.setRecreateNewWindowEnabled( true );
                                client.virtualClientAction( UNDEFINED_WINDOW, VirtualClientAction.RECREATE_NEW_CHILD_WINDOW );
                            }
                            else if ( waitInterval >= MAX_WAIT_WINDOW_CREATED_FIRST_INTERVAL && waitInterval < MAX_WAIT_INTERVAL
                                && client.getDialogAndNormalAndSplashWindowsCounter() >= ONLY_ONE_WINDOW && virtualUtil.isNotDrawingStarted() && virtualUtil.isNotRecreateNewWindowEnabled() )
                            {
                                if ( client.getDialogWindowsCounter() >= ONLY_ONE_WINDOW || client.getSplashWindowsCounter() >= ONLY_ONE_WINDOW )
                                {
                                    virtualUtil.setDrawingStarted( true );
                                    windowCreated( true );
                                }
                                else
                                {
                                    virtualUtil.setRecreateNewWindowEnabled( true );
                                    virtualUtil.setRecreateNewWindowAgainEnabled( false );
                                    client.virtualClientAction( UNDEFINED_WINDOW, VirtualClientAction.RECREATE_NEW_WINDOW );
                                }
                            }
                            else if ( waitInterval >= MAX_WAIT_WINDOW_CREATED_SECOND_INTERVAL && waitInterval < MAX_WAIT_INTERVAL
                                && client.getDialogAndNormalAndSplashWindowsCounter() >= ONLY_ONE_WINDOW && virtualUtil.isNotDrawingStarted() && virtualUtil.isNotRecreateNewWindowAgainEnabled() )
                            {
                                if ( client.getDialogWindowsCounter() >= ONLY_ONE_WINDOW || client.getSplashWindowsCounter() >= ONLY_ONE_WINDOW )
                                {
                                    virtualUtil.setDrawingStarted( true );
                                    windowCreated( true );
                                }
                                else
                                {
                                    virtualUtil.setRecreateNewWindowAgainEnabled( true );
                                    client.virtualClientAction( UNDEFINED_WINDOW, VirtualClientAction.RECREATE_NEW_WINDOW );
                                }
                            }
                            else if ( waitInterval >= MAX_WAIT_WINDOW_CREATED_FIRST_INTERVAL && waitInterval < MAX_WAIT_INTERVAL && client.getNormalWindowsCounter() > ONLY_ONE_WINDOW
                                && client.isNotWindowCreated() )
                            {
                                client.getSender()
                                      .onSend( new ClientFocusPacket( client.getLastWindowId() ) );
                                windowCreated( true );
                            }
                            else if ( waitInterval >= MAX_WAIT_INTERVAL && client.isNotWindowCreated() )
                            {
                                if ( dialogConnectingUtil.isProgressBarVisible() )
                                {
                                    canvasDiv.getUI()
                                             .ifPresent( ui -> ui.getSession()
                                                                 .lock() );
                                    canvasDiv.getUI()
                                             .ifPresent( ui -> ui.access( () -> {
                                                 dialogConnectingUtil.setProgressValue( DialogConnectingUtil.PROGRESS_STOP );
                                                 ui.push();
                                             } ) );
                                    canvasDiv.getUI()
                                             .ifPresent( ui -> ui.getSession()
                                                                 .unlock() );
                                }
                                connector.setErrorMessage( virtualUtil.getTranslation( Translation.APPLICATION_ERROR_EXECUTE ) );
                                connector.stopApplication();
                            }
                        }

                        // progress
                        if ( waitInterval >= PROGRESS_SHOW_INTERVAL && dialogConnectingUtil.isOpened() )
                        {
                            canvasDiv.getUI()
                                     .ifPresent( ui -> ui.getSession()
                                                         .lock() );
                            canvasDiv.getUI()
                                     .ifPresent( ui -> ui.access( () -> {
                                         double progress = (double) waitInterval / MAX_WAIT_INTERVAL * DialogConnectingUtil.PROGRESS_STOP;
                                         dialogConnectingUtil.setProgressBarVisible( true );
                                         dialogConnectingUtil.setProgressValue( progress );
                                         ui.push();
                                     } ) );
                            canvasDiv.getUI()
                                     .ifPresent( ui -> ui.getSession()
                                                         .unlock() );

                            if ( dialogConnectingUtil.getProgressValue() >= DialogConnectingUtil.PROGRESS_STOP )
                            {
                                TimeUnit.MILLISECONDS.sleep( SLEEP_STEP_INTERVAL );
                                connector.setErrorMessage( virtualUtil.getTranslation( Translation.APPLICATION_ERROR_EXECUTE ) );
                                virtualUtil.stopConnectorService();
                            }
                        }
                    }
                }

                // process the error message
                if ( Objects.nonNull( connector.getErrorMessage() ) )
                {
                    canvasDiv.getUI()
                             .ifPresent( ui -> ui.access( () -> {
                                 String message = virtualUtil.getTranslation( Translation.APPLICATION_ERROR_CONNECT );
                                 if ( StringUtils.equals( connector.getErrorMessage(), virtualUtil.getTranslation( Translation.APPLICATION_ERROR_CONNECTION ) ) )
                                 {
                                     message = virtualUtil.getTranslation( Translation.APPLICATION_ERROR_CONNECTION );
                                     dialogConnectingUtil.setOpened( true );
                                 }
                                 else if ( StringUtils.equals( connector.getErrorMessage(), virtualUtil.getTranslation( Translation.APPLICATION_ERROR_EXECUTE ) ) )
                                 {
                                     message = virtualUtil.getTranslation( Translation.APPLICATION_ERROR_EXECUTE );
                                 }

                                 if ( virtualUtil.isNotApplicationTesting() )
                                 {
                                     dialogConnectingUtil.setButtonsVisible( true );
                                     dialogConnectingUtil.setMessage( message );
                                 }
                                 else
                                 {
                                     windowCreated( false );
                                 }
                                 ui.push();
                             } ) );

                    // fatal error so stop instance of the application
                    connector.resetErrorMessage();
                    connector.stopServer();
                }
            }
            catch ( InterruptedException exception )
            {
                LOGGER.warn( LoggerUtil.THREAD_INTERRUPTED, exception );
                Thread.currentThread()
                      .interrupt();
            }

            if ( virtualUtil.isNotBrowserOpened() )
            {
                // browser is closed while application still running, so disconnect from the server
                connector.disconnect();
            }
            else
            {
                LOGGER.info( CONNECTION_CLOSED );
            }

            // stop the server
            if ( connector.haveToStopServer() && Objects.nonNull( database ) && Objects.nonNull( userSessionData ) )
            {
                StopServerRunner stopServerRunner = new StopServerRunner( database, userSessionData.getId(), userSessionData.getServerData(), userSessionData.getPort(), virtualUtil );
                virtualUtil.startRunnerService( stopServerRunner );

                // instance is disconnected by server error
                if ( dialogConnectingUtil.isOpened() || dialogConnectingUtil.haveToShowDialog() )
                {
                    Client client = connector.getClient();
                    canvasDiv.getUI()
                             .ifPresent( ui -> ui.access( () -> {
                                 if ( client.isDisconnectedByServerError() )
                                 {
                                     dialogConnectingUtil.setMessage( virtualUtil.getTranslation( Translation.APPLICATION_ERROR_INSTANCE ) );
                                 }
                                 else if ( client.isDisconnectedBySecurityError() )
                                 {
                                     dialogConnectingUtil.setMessage( virtualUtil.getTranslation( Translation.APPLICATION_RUN_ERROR_INSTANCE ) );
                                     dialogConnectingUtil.setOpened( true );
                                 }
                                 else
                                 {
                                     client.setDisconnectedByServerError( true );
                                     dialogConnectingUtil.setMessage( virtualUtil.getTranslation( Translation.APPLICATION_ERROR_CONNECT ) );
                                 }
                                 dialogConnectingUtil.setButtonsVisible( true );
                                 dialogConnectingUtil.addDetachListener( event -> client.virtualClientAction( UNDEFINED_WINDOW, VirtualClientAction.CLOSE_ALL_WINDOWS ) );
                                 ui.push();
                             } ) );
                }
            }
            else if ( connector.haveToStopServer() && Objects.nonNull( userSessionData ) )
            {
                ServerData serverData = userSessionData.getServerData();
                LOGGER.info( CANNOT_STOP_SERVER, serverData.getAddress(), serverData.getSshPort(), userSessionData.getPort() );
            }
            virtualUtil.setConnectionToServerActive( false );

            // reset wait intervals
            resetWaitIntervals();
        }
    }

}
