/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.ui.applicationgroups;

import java.util.ArrayList;
import java.util.EnumMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.stream.Collectors;
import java.util.stream.LongStream;

import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dvirtual.database.Database;
import org.dussan.vaadin.dvirtual.database.columns.ApplicationColumns;
import org.dussan.vaadin.dvirtual.database.data.ApplicationData;
import org.dussan.vaadin.dvirtual.database.data.ApplicationsGroupData;
import org.dussan.vaadin.dvirtual.database.data.ServersGroupData;
import org.dussan.vaadin.dvirtual.i18n.Translation;
import org.dussan.vaadin.dvirtual.utils.VirtualUtil;
import org.dussan.vaadin.dvirtual.utils.dialog.DialogEditUtil;
import org.dussan.vaadin.dvirtual.utils.dialog.DialogInfoUtil;

import com.vaadin.flow.component.button.Button;
import com.vaadin.flow.component.checkbox.Checkbox;
import com.vaadin.flow.component.combobox.ComboBox;
import com.vaadin.flow.component.formlayout.FormLayout;
import com.vaadin.flow.component.formlayout.FormLayout.ResponsiveStep;
import com.vaadin.flow.component.formlayout.FormLayout.ResponsiveStep.LabelsPosition;
import com.vaadin.flow.component.grid.Grid;
import com.vaadin.flow.component.grid.Grid.SelectionMode;
import com.vaadin.flow.component.grid.HeaderRow;
import com.vaadin.flow.component.html.Div;
import com.vaadin.flow.component.html.Label;
import com.vaadin.flow.component.icon.Icon;
import com.vaadin.flow.component.icon.VaadinIcon;
import com.vaadin.flow.component.orderedlayout.FlexComponent.Alignment;
import com.vaadin.flow.component.orderedlayout.FlexComponent.JustifyContentMode;
import com.vaadin.flow.component.orderedlayout.HorizontalLayout;
import com.vaadin.flow.component.orderedlayout.VerticalLayout;
import com.vaadin.flow.component.textfield.TextField;
import com.vaadin.flow.data.binder.Binder;
import com.vaadin.flow.data.provider.DataProvider;
import com.vaadin.flow.data.provider.SortDirection;
import com.vaadin.flow.data.provider.SortOrder;
import com.vaadin.flow.data.renderer.ComponentRenderer;
import com.vaadin.flow.data.validator.StringLengthValidator;
import com.vaadin.flow.data.value.ValueChangeMode;

public abstract class ApplicationsGroupEditDialog
    extends DialogEditUtil
{

    private static final long serialVersionUID = 4242484379277525032L;

    private static final int FORM_COLUMNS = 1;

    private static final int NO_LIMITS = Integer.MAX_VALUE;

    private static final int VALUE_CHANGE_TIMEOUT = 500;

    private static final int VISIBLE_ROWS = 50;

    private static final String ATTRIBUTE_CLASS_BUTTON_VALUE = "primary";

    private static final String ATTRIBUTE_THEME = "theme";

    private static final String ATTRIBUTE_THEME_VALUE = "dark";

    private static final String CLASS_DIALOG = "edit-dialog";

    private static final String CLASS_GRID_CELL_ACTIVE = "active";

    private static final String CLASS_GRID_CELL_NOT_ACTIVE = StringUtils.EMPTY;

    private static final String CLASS_GRID_CELL_THEME = "grid-cell-theme";

    private static final String CLASS_GRID_GROUPS = "grid-groups";

    private static final String CLASS_GRID_GROUPS_ICON = "grid-groups-icon";

    private static final String CLASS_GRID_GROUPS_LABEL = "grid-groups-label";

    private static final String CLASS_GRID_LABEL = "grid-label";

    private static final String FORM_MINIMAL_WIDTH = "0";

    private static final String STYLE_DISPLAY = "display";

    private static final String STYLE_DISPLAY_GRID_VALUE = "grid";

    private static final String STYLE_GRID_COMBOBOX = "grid-combobox";

    private static final String STYLE_HEIGHT = "height";

    private static final String STYLE_HEIGHT_16PX_VALUE = "16px";

    private static final String STYLE_MARGIN_BOTTOM = "margin-bottom";

    private static final String STYLE_MARGIN_LEFT = "margin-left";

    private static final String STYLE_MARGIN_RIGHT = "margin-right";

    private static final String STYLE_MARGIN_VALUE = "0px";

    private static final String STYLE_PADDING_14PX_VALUE = "14px";

    private static final String STYLE_PADDING_BOTTOM = "padding-bottom";

    private static final String STYLE_PADDING_LEFT = "padding-left";

    private static final String STYLE_PADDING_RIGHT = "padding-right";

    private static final String STYLE_PADDING_TOP = "padding-top";

    private static final String STYLE_PADDING_VALUE = "0px";

    private static final String STYLE_TEXT_ALIGN = "text-align";

    private static final String STYLE_TEXT_ALIGN_VALUE = "center";

    private static final String STYLE_WIDTH = "width";

    private static final String STYLE_WIDTH_100PERCENT_VALUE = "100%";

    private static final String STYLE_WIDTH_130PX_VALUE = "130px";

    private static final String STYLE_WIDTH_150PX_VALUE = "150px";

    private static final String STYLE_WIDTH_16PX_VALUE = "16px";

    private static final String STYLE_WIDTH_50PX_VALUE = "50px";

    private transient List<ApplicationData> selectedApplicationDatas = null;

    private transient Map<ApplicationColumns, String> filterMap = null;

    private Checkbox enabledCheckbox = null;

    private Checkbox sameInstanceCheckbox = null;

    private Checkbox selectAllCheckbox = null;

    private DataProvider<ApplicationData, Void> dataProvider = null;

    private Grid<ApplicationData> applicationGrid = null;

    private TextField nameTextField = null;

    private TextField noteTextField = null;

    public ApplicationsGroupEditDialog( long applicationsGroupId, ApplicationsGroupData applicationsGroupData, Database database, VirtualUtil virtualUtil )
    {
        filterMap = new EnumMap<>( ApplicationColumns.class );
        selectedApplicationDatas = applicationsGroupId == ApplicationsGroupData.NEW_APPLICATIONS_GROUP ? new ArrayList<>() : applicationsGroupData.getApplicationDatas();

        // close/save buttons
        Button closeButton = new Button( virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) );
        closeButton.getElement()
                   .setAttribute( ATTRIBUTE_THEME, ATTRIBUTE_THEME_VALUE );
        Button saveButton = new Button( virtualUtil.getTranslation( Translation.BUTTON_SAVE ) );
        saveButton.setEnabled( false );
        saveButton.addClassName( ATTRIBUTE_CLASS_BUTTON_VALUE );
        saveButton.getElement()
                  .setAttribute( ATTRIBUTE_THEME, ATTRIBUTE_THEME_VALUE );
        saveButton.getStyle()
                  .set( STYLE_MARGIN_LEFT, STYLE_MARGIN_VALUE );

        // buttons layout
        HorizontalLayout buttonsLayout = new HorizontalLayout( saveButton, closeButton );
        buttonsLayout.setJustifyContentMode( JustifyContentMode.CENTER );
        buttonsLayout.setMargin( true );
        buttonsLayout.setPadding( true );
        buttonsLayout.setSpacing( true );
        buttonsLayout.setWidth( STYLE_WIDTH_100PERCENT_VALUE );
        buttonsLayout.getStyle()
                     .set( STYLE_MARGIN_BOTTOM, STYLE_MARGIN_VALUE )
                     .set( STYLE_MARGIN_LEFT, STYLE_MARGIN_VALUE )
                     .set( STYLE_MARGIN_RIGHT, STYLE_MARGIN_VALUE )
                     .set( STYLE_PADDING_BOTTOM, STYLE_PADDING_VALUE )
                     .set( STYLE_PADDING_LEFT, STYLE_PADDING_VALUE )
                     .set( STYLE_PADDING_RIGHT, STYLE_PADDING_VALUE );

        // name
        nameTextField = new TextField( virtualUtil.getTranslation( Translation.FIELD_NAME ), StringUtils.stripToEmpty( applicationsGroupData.getName() ),
                                       event -> saveButton.setEnabled( isAllRequiredValuesNonEmpty() && !StringUtils.equals( event.getValue(), event.getOldValue() ) ) );
        nameTextField.setAutofocus( true );
        nameTextField.setRequiredIndicatorVisible( true );
        nameTextField.setValueChangeMode( ValueChangeMode.LAZY );
        nameTextField.setValueChangeTimeout( VALUE_CHANGE_TIMEOUT );
        nameTextField.getStyle()
                     .set( STYLE_PADDING_TOP, STYLE_PADDING_VALUE );

        // same instance
        sameInstanceCheckbox =
            new Checkbox( virtualUtil.getTranslation( Translation.FIELD_APPLICATIONS_GROUP_INSTANCE ),
                          applicationsGroupId == ApplicationsGroupData.NEW_APPLICATIONS_GROUP ? ApplicationsGroupData.DEFAULT_SAME_INSTANCE : applicationsGroupData.isSameInstance() );
        sameInstanceCheckbox.addValueChangeListener( event -> saveButton.setEnabled( isAllRequiredValuesNonEmpty() && !Objects.equals( event.getValue(), event.getOldValue() ) ) );
        // enabled
        enabledCheckbox = new Checkbox( virtualUtil.getTranslation( Translation.FIELD_ENABLED ),
                                        applicationsGroupId == ApplicationsGroupData.NEW_APPLICATIONS_GROUP ? ApplicationsGroupData.DEFAULT_ENABLED : applicationsGroupData.isEnabled() );
        enabledCheckbox.addValueChangeListener( event -> saveButton.setEnabled( isAllRequiredValuesNonEmpty() && !Objects.equals( event.getValue(), event.getOldValue() ) ) );
        // name layout: name, enabled
        HorizontalLayout nameLayout = new HorizontalLayout( nameTextField, sameInstanceCheckbox, enabledCheckbox );
        nameLayout.expand( nameTextField );
        nameLayout.setDefaultVerticalComponentAlignment( Alignment.BASELINE );
        nameLayout.setPadding( false );
        nameLayout.setSpacing( true );
        nameLayout.setWidth( STYLE_WIDTH_100PERCENT_VALUE );

        // note
        noteTextField = new TextField( virtualUtil.getTranslation( Translation.FIELD_NOTE ), StringUtils.stripToEmpty( applicationsGroupData.getNote() ),
                                       event -> saveButton.setEnabled( isAllRequiredValuesNonEmpty() && !StringUtils.equals( event.getValue(), event.getOldValue() ) ) );
        noteTextField.setValueChangeMode( ValueChangeMode.LAZY );
        noteTextField.setValueChangeTimeout( VALUE_CHANGE_TIMEOUT );
        noteTextField.setWidth( STYLE_WIDTH_100PERCENT_VALUE );
        noteTextField.getStyle()
                     .set( STYLE_PADDING_BOTTOM, STYLE_PADDING_14PX_VALUE );

        // application data provider
        dataProvider = DataProvider.fromCallbacks( query -> {
            int offset = query.getOffset();
            int limit = query.getLimit();

            List<String> orders = new ArrayList<>();
            for ( SortOrder<String> order : query.getSortOrders() )
            {
                orders.add( order.getDirection()
                                 .equals( SortDirection.DESCENDING ) ? Database.ORDER_DESCENDING + order.getSorted() : order.getSorted() );
            }

            return new ArrayList<>( database.getApplications( offset, limit, filterMap, orders ) ).stream();
        }, query -> {
            long[] selectedApplicationIds = selectedApplicationDatas.stream()
                                                                    .mapToLong( ApplicationData::getId )
                                                                    .toArray();
            int filteredSelectedAplicationsSize = //
                database.getApplications( 0, NO_LIMITS, filterMap, null )
                        .stream()
                        .filter( applicationData -> LongStream.of( selectedApplicationIds )
                                                              .anyMatch( selectedApplicationId -> selectedApplicationId == applicationData.getId() ) )
                        .collect( Collectors.toList() )
                        .size();

            Long applicationRows = database.getApplicationRows( filterMap );
            selectAllCheckbox.setValue( filteredSelectedAplicationsSize == applicationRows && applicationRows > 0 );

            return applicationRows.intValue();
        } );

        // application
        applicationGrid = new Grid<>();
        applicationGrid.setDataProvider( dataProvider );
        applicationGrid.setId( CLASS_GRID_CELL_THEME );
        applicationGrid.setPageSize( VISIBLE_ROWS );
        applicationGrid.setSelectionMode( SelectionMode.NONE );
        applicationGrid.getStyle()
                       .set( STYLE_DISPLAY, STYLE_DISPLAY_GRID_VALUE )
                       .set( STYLE_WIDTH, STYLE_WIDTH_100PERCENT_VALUE );

        // application: selected
        selectAllCheckbox = new Checkbox();
        selectAllCheckbox.addClickListener( event -> {
            selectedApplicationDatas.clear();
            if ( Boolean.TRUE.equals( event.getSource()
                                           .getValue() ) )
            {
                selectedApplicationDatas.addAll( database.getApplications( 0, NO_LIMITS, filterMap, null ) );
            }
            saveButton.setEnabled( isAllRequiredValuesNonEmpty() );
            dataProvider.refreshAll();
        } );
        applicationGrid.addColumn( new ComponentRenderer<>( applicationData -> {
            Checkbox checkbox = new Checkbox();
            checkbox.setValue( selectedApplicationDatas.stream()
                                                       .anyMatch( selectedApplicationData -> Objects.equals( selectedApplicationData.getId(), applicationData.getId() ) ) );
            checkbox.addClickListener( event -> {
                if ( Boolean.TRUE.equals( event.getSource()
                                               .getValue() ) )
                {
                    selectedApplicationDatas.add( applicationData );
                }
                else
                {
                    selectedApplicationDatas.removeIf( selectedApplicationData -> Objects.equals( selectedApplicationData.getId(), applicationData.getId() ) );
                }
                saveButton.setEnabled( isAllRequiredValuesNonEmpty() );
                dataProvider.refreshAll();
            } );
            return checkbox;
        } ) )
                       .setFlexGrow( 0 )
                       .setFrozen( true )
                       .setKey( ApplicationColumns.SELECTED.getName() )
                       .setResizable( false )
                       .setSortable( false )
                       .setWidth( STYLE_WIDTH_50PX_VALUE )
                       .setHeader( selectAllCheckbox );

        // application: name
        Grid.Column<ApplicationData> nameColumn = //
            applicationGrid.addColumn( ApplicationData::getName )
                           .setFlexGrow( 1 )
                           .setFrozen( true )
                           .setKey( ApplicationColumns.NAME.getName() )
                           .setResizable( true )
                           .setSortable( true )
                           .setSortProperty( ApplicationColumns.NAME.getName() )
                           .setWidth( STYLE_WIDTH_150PX_VALUE )
                           .setHeader( virtualUtil.getTranslation( ApplicationColumns.NAME.getValueForTranslation() ) );

        // application: note
        Grid.Column<ApplicationData> noteColumn = //
            applicationGrid.addColumn( ApplicationData::getNote )
                           .setFlexGrow( 2 )
                           .setKey( ApplicationColumns.NOTE.getName() )
                           .setResizable( true )
                           .setSortable( true )
                           .setSortProperty( ApplicationColumns.NOTE.getName() )
                           .setHeader( virtualUtil.getTranslation( ApplicationColumns.NOTE.getValueForTranslation() ) );

        // application: servers group
        Grid.Column<ApplicationData> serversGroupColumn = //
            applicationGrid.addColumn( new ComponentRenderer<>( Div::new, ( serverGroupsDiv, applicationData ) -> {
                List<ServersGroupData> serversGroupDatas = applicationData.getServersGroupDatas();
                serversGroupDatas.stream()
                                 .filter( Objects::nonNull )
                                 .forEach( serversGroupData -> {
                                     Label label = new Label( serversGroupData.getName() );
                                     label.addClassName( CLASS_GRID_GROUPS_LABEL );
                                     Icon icon = serversGroupData.isEnabled() ? new Icon( VaadinIcon.CHECK ) : new Icon( VaadinIcon.CLOSE );
                                     icon.addClassName( CLASS_GRID_GROUPS_ICON );

                                     Div serversGroupDiv = new Div();
                                     serversGroupDiv.addClassName( CLASS_GRID_GROUPS );
                                     serversGroupDiv.add( icon );
                                     serversGroupDiv.add( label );
                                     serverGroupsDiv.add( serversGroupDiv );
                                 } );
            } ) )
                           .setFlexGrow( 1 )
                           .setKey( ApplicationColumns.SERVERS_GROUP.getName() )
                           .setResizable( true )
                           .setSortable( false )
                           .setHeader( virtualUtil.getTranslation( ApplicationColumns.SERVERS_GROUP.getValueForTranslation() ) );

        // application: one instance
        Grid.Column<ApplicationData> oneInstanceColumn = //
            applicationGrid.addColumn( new ComponentRenderer<>( Div::new, ( div, applicationData ) -> {
                Icon icon = applicationData.isOneInstance() ? new Icon( VaadinIcon.CHECK ) : new Icon( VaadinIcon.CLOSE );
                icon.getStyle()
                    .set( STYLE_HEIGHT, STYLE_HEIGHT_16PX_VALUE )
                    .set( STYLE_WIDTH, STYLE_WIDTH_16PX_VALUE );
                div.add( icon );
                div.getStyle()
                   .set( STYLE_TEXT_ALIGN, STYLE_TEXT_ALIGN_VALUE );
            } ) )
                           .setFlexGrow( 0 )
                           .setKey( ApplicationColumns.ONE_INSTANCE.getName() )
                           .setResizable( true )
                           .setSortable( true )
                           .setSortProperty( ApplicationColumns.ONE_INSTANCE.getName() )
                           .setWidth( STYLE_WIDTH_130PX_VALUE )
                           .setHeader( virtualUtil.getTranslation( ApplicationColumns.ONE_INSTANCE.getValueForTranslation() ) );

        // application: enabled
        Grid.Column<ApplicationData> enabledColumn = //
            applicationGrid.addColumn( new ComponentRenderer<>( Div::new, ( div, applicationData ) -> {
                Icon icon = applicationData.isEnabled() ? new Icon( VaadinIcon.CHECK ) : new Icon( VaadinIcon.CLOSE );
                icon.getStyle()
                    .set( STYLE_HEIGHT, STYLE_HEIGHT_16PX_VALUE )
                    .set( STYLE_WIDTH, STYLE_WIDTH_16PX_VALUE );
                div.add( icon );
                div.getStyle()
                   .set( STYLE_TEXT_ALIGN, STYLE_TEXT_ALIGN_VALUE );
            } ) )
                           .setFlexGrow( 0 )
                           .setKey( ApplicationColumns.ENABLED.getName() )
                           .setResizable( true )
                           .setSortable( true )
                           .setSortProperty( ApplicationColumns.ENABLED.getName() )
                           .setWidth( STYLE_WIDTH_130PX_VALUE )
                           .setHeader( virtualUtil.getTranslation( ApplicationColumns.ENABLED.getValueForTranslation() ) );

        Label applicationGridLabel = new Label( virtualUtil.getTranslation( Translation.APPLICATIONS ) );
        applicationGridLabel.setFor( virtualUtil.getTranslation( Translation.APPLICATIONS ) );
        applicationGridLabel.addClassName( CLASS_GRID_LABEL );

        VerticalLayout applicationGridLayout = new VerticalLayout( applicationGridLabel, applicationGrid );
        applicationGridLayout.setMargin( false );
        applicationGridLayout.setPadding( false );
        applicationGridLayout.setSpacing( false );
        applicationGridLayout.setSizeUndefined();
        applicationGridLayout.setWidth( STYLE_WIDTH_100PERCENT_VALUE );

        // legend
        Label legendSameInstance = new Label( virtualUtil.getTranslation( Translation.APPLICATIONS_GROUP_INSTANCE ) );
        VerticalLayout legendLayout = new VerticalLayout( legendSameInstance );
        legendLayout.setSpacing( false );
        legendLayout.setWidth( STYLE_WIDTH_100PERCENT_VALUE );
        legendLayout.getStyle()
                    .set( STYLE_PADDING_BOTTOM, STYLE_PADDING_VALUE )
                    .set( STYLE_PADDING_LEFT, STYLE_PADDING_VALUE )
                    .set( STYLE_PADDING_RIGHT, STYLE_PADDING_VALUE );

        Binder<ApplicationsGroupData> applicationsGroupBinder = new Binder<>();
        applicationsGroupBinder.forField( nameTextField )
                               .withValidator( new StringLengthValidator( virtualUtil.getTranslation( Translation.DATA_VALUE ), 1, null ) )
                               .bind( ApplicationsGroupData::getName, ApplicationsGroupData::setName );
        applicationsGroupBinder.bind( noteTextField, ApplicationsGroupData::getNote, ApplicationsGroupData::setNote );
        applicationsGroupBinder.bind( sameInstanceCheckbox, ApplicationsGroupData::isSameInstance, ApplicationsGroupData::setSameInstance );
        applicationsGroupBinder.bind( enabledCheckbox, ApplicationsGroupData::isEnabled, ApplicationsGroupData::setEnabled );

        closeButton.addClickListener( event -> close() );
        saveButton.addClickListener( event -> {
            ApplicationsGroupData data = new ApplicationsGroupData();
            if ( applicationsGroupBinder.writeBeanIfValid( data ) )
            {
                data.setId( applicationsGroupId );
                data.setApplicationDatas( selectedApplicationDatas );
                savingData( data );
            }
            else
            {
                new DialogInfoUtil( virtualUtil.getTranslation( Translation.DATA_ERROR_SAVE ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
            }
        } );

        // filter: name
        HeaderRow filterRow = applicationGrid.appendHeaderRow();
        TextField nameFilter = new TextField();
        nameFilter.setPlaceholder( virtualUtil.getTranslation( Translation.FIELD_FILTER ) );
        nameFilter.setSizeFull();
        nameFilter.setValueChangeMode( ValueChangeMode.LAZY );
        nameFilter.setValueChangeTimeout( VALUE_CHANGE_TIMEOUT );
        nameFilter.addValueChangeListener( event -> {
            if ( nameFilter.isEmpty() )
            {
                filterMap.remove( ApplicationColumns.NAME );
            }
            else
            {
                filterMap.put( ApplicationColumns.NAME, nameFilter.getValue() );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( nameColumn )
                 .setComponent( nameFilter );

        // filter: note
        TextField noteFilter = new TextField();
        noteFilter.setPlaceholder( virtualUtil.getTranslation( Translation.FIELD_FILTER ) );
        noteFilter.setSizeFull();
        noteFilter.setValueChangeMode( ValueChangeMode.LAZY );
        noteFilter.setValueChangeTimeout( VALUE_CHANGE_TIMEOUT );
        noteFilter.addValueChangeListener( event -> {
            if ( noteFilter.isEmpty() )
            {
                filterMap.remove( ApplicationColumns.NOTE );
            }
            else
            {
                filterMap.put( ApplicationColumns.NOTE, noteFilter.getValue() );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( noteColumn )
                 .setComponent( noteFilter );

        // filter: servers group
        TextField serversGroupFilter = new TextField();
        serversGroupFilter.setPlaceholder( virtualUtil.getTranslation( Translation.FIELD_FILTER ) );
        serversGroupFilter.setSizeFull();
        serversGroupFilter.setValueChangeMode( ValueChangeMode.LAZY );
        serversGroupFilter.setValueChangeTimeout( VALUE_CHANGE_TIMEOUT );
        serversGroupFilter.addValueChangeListener( event -> {
            if ( serversGroupFilter.isEmpty() )
            {
                filterMap.remove( ApplicationColumns.SERVERS_GROUP );
            }
            else
            {
                filterMap.put( ApplicationColumns.SERVERS_GROUP, serversGroupFilter.getValue() );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( serversGroupColumn )
                 .setComponent( serversGroupFilter );

        // filter: one instance
        String filterAll = virtualUtil.getTranslation( Translation.DATA_FILTER_ALL );
        String filterNo = virtualUtil.getTranslation( Translation.DATA_FILTER_NO );
        String filterYes = virtualUtil.getTranslation( Translation.DATA_FILTER_YES );
        ComboBox<String> oneInstanceFilter = new ComboBox<>();
        oneInstanceFilter.setAllowCustomValue( false );
        oneInstanceFilter.setClassName( STYLE_GRID_COMBOBOX );
        oneInstanceFilter.setItems( filterAll, filterYes, filterNo );
        oneInstanceFilter.setValue( filterAll );
        oneInstanceFilter.setSizeFull();
        oneInstanceFilter.addValueChangeListener( event -> {
            if ( StringUtils.equals( oneInstanceFilter.getValue(), filterYes ) )
            {
                filterMap.put( ApplicationColumns.ONE_INSTANCE, Boolean.TRUE.toString() );
            }
            else if ( StringUtils.equals( oneInstanceFilter.getValue(), filterNo ) )
            {
                filterMap.put( ApplicationColumns.ONE_INSTANCE, Boolean.FALSE.toString() );
            }
            else
            {
                filterMap.remove( ApplicationColumns.ONE_INSTANCE );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( oneInstanceColumn )
                 .setComponent( oneInstanceFilter );

        // filter: enabled
        ComboBox<String> enabledFilter = new ComboBox<>();
        enabledFilter.setAllowCustomValue( false );
        enabledFilter.setClassName( STYLE_GRID_COMBOBOX );
        enabledFilter.setItems( filterAll, filterYes, filterNo );
        enabledFilter.setValue( filterAll );
        enabledFilter.setSizeFull();
        enabledFilter.addValueChangeListener( event -> {
            if ( StringUtils.equals( enabledFilter.getValue(), filterYes ) )
            {
                filterMap.put( ApplicationColumns.ENABLED, Boolean.TRUE.toString() );
            }
            else if ( StringUtils.equals( enabledFilter.getValue(), filterNo ) )
            {
                filterMap.put( ApplicationColumns.ENABLED, Boolean.FALSE.toString() );
            }
            else
            {
                filterMap.remove( ApplicationColumns.ENABLED );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( enabledColumn )
                 .setComponent( enabledFilter );

        // form layout
        setFormLayout( new FormLayout( nameLayout, noteTextField, applicationGridLayout, legendLayout, buttonsLayout ) );
        getFormLayout().setClassName( CLASS_DIALOG );
        getFormLayout().setResponsiveSteps( new ResponsiveStep( FORM_MINIMAL_WIDTH, FORM_COLUMNS, LabelsPosition.TOP ) );
        getFormLayout().setWidth( STYLE_WIDTH_100PERCENT_VALUE );
        add( getFormLayout() );

        // set active/selected rows
        applicationGrid.getColumnByKey( ApplicationColumns.ENABLED.getName() )
                       .setClassNameGenerator( applicationData -> getCellClassName( applicationData.getId() ) );
        applicationGrid.getColumnByKey( ApplicationColumns.NAME.getName() )
                       .setClassNameGenerator( applicationData -> getCellClassName( applicationData.getId() ) );
        applicationGrid.getColumnByKey( ApplicationColumns.NOTE.getName() )
                       .setClassNameGenerator( applicationData -> getCellClassName( applicationData.getId() ) );
        applicationGrid.getColumnByKey( ApplicationColumns.ONE_INSTANCE.getName() )
                       .setClassNameGenerator( applicationData -> getCellClassName( applicationData.getId() ) );
        applicationGrid.getColumnByKey( ApplicationColumns.SELECTED.getName() )
                       .setClassNameGenerator( applicationData -> getCellClassName( applicationData.getId() ) );
        applicationGrid.getColumnByKey( ApplicationColumns.SERVERS_GROUP.getName() )
                       .setClassNameGenerator( applicationData -> getCellClassName( applicationData.getId() ) );
    }

    public abstract void savingData( ApplicationsGroupData applicationsGroupData );

    private boolean isAllRequiredValuesNonEmpty()
    {
        return ( StringUtils.isNotBlank( nameTextField.getValue() ) && CollectionUtils.isNotEmpty( selectedApplicationDatas ) );
    }

    private String getCellClassName( Long applicationDataId )
    {
        if ( selectedApplicationDatas.stream()
                                     .anyMatch( selectedApplicationData -> Objects.equals( selectedApplicationData.getId(), applicationDataId ) ) )
        {
            return CLASS_GRID_CELL_ACTIVE;
        }
        return CLASS_GRID_CELL_NOT_ACTIVE;
    }

}
