/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.ui.bars.applications;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dvirtual.client.Command;
import org.dussan.vaadin.dvirtual.database.data.ServerData;
import org.dussan.vaadin.dvirtual.i18n.Translation;
import org.dussan.vaadin.dvirtual.utils.VirtualUtil;
import org.dussan.vaadin.dvirtual.utils.dialog.common.DialogUtil;

import com.vaadin.flow.component.AttachEvent;
import com.vaadin.flow.component.DetachEvent;
import com.vaadin.flow.component.button.Button;
import com.vaadin.flow.component.formlayout.FormLayout;
import com.vaadin.flow.component.formlayout.FormLayout.ResponsiveStep;
import com.vaadin.flow.component.formlayout.FormLayout.ResponsiveStep.LabelsPosition;
import com.vaadin.flow.component.grid.Grid;
import com.vaadin.flow.component.grid.HeaderRow;
import com.vaadin.flow.component.grid.Grid.SelectionMode;
import com.vaadin.flow.component.orderedlayout.HorizontalLayout;
import com.vaadin.flow.component.orderedlayout.VerticalLayout;
import com.vaadin.flow.component.orderedlayout.FlexComponent.JustifyContentMode;
import com.vaadin.flow.component.textfield.TextField;
import com.vaadin.flow.data.provider.ListDataProvider;
import com.vaadin.flow.data.value.ValueChangeMode;

public abstract class SelectServerDialog
    extends DialogUtil
{

    private static final long serialVersionUID = 2888747991258404837L;

    private static final int FORM_COLUMNS = 1;

    private static final int GRID_PAGE_SIZE = 50;

    private static final String ATTRIBUTE_CLASS_BUTTON_VALUE = "primary";

    private static final String ATTRIBUTE_THEME = "theme";

    private static final String ATTRIBUTE_THEME_VALUE = "dark";

    private static final String DIALOG_CLASS = "edit-dialog";

    private static final String DIALOG_HEIGHT = "auto";

    private static final String DIALOG_WIDTH = "auto";

    private static final String FORM_MINIMAL_WIDTH = "0";

    private static final String STYLE_DISPLAY = "display";

    private static final String STYLE_DISPLAY_VALUE = "grid";

    private static final String STYLE_MARGIN_BOTTOM = "margin-bottom";

    private static final String STYLE_MARGIN_LEFT = "margin-left";

    private static final String STYLE_MARGIN_RIGHT = "margin-right";

    private static final String STYLE_MARGIN_0PX_VALUE = "0px";

    private static final String STYLE_MARGIN_1PX_VALUE = "1px";

    private static final String STYLE_PADDING_BOTTOM = "padding-bottom";

    private static final String STYLE_PADDING_LEFT = "padding-left";

    private static final String STYLE_PADDING_RIGHT = "padding-right";

    private static final String STYLE_PADDING_VALUE = "0px";

    private static final String STYLE_WIDTH_100PERCENT_VALUE = "100%";

    private static final String STYLE_WIDTH_150PX_VALUE = "150px";

    private transient List<ServerData> serverDatas = null;

    private FormLayout formLayout = null;

    private Grid<ServerData> serverGrid = null;

    private ListDataProvider<ServerData> dataProvider = null;

    private SelectServerDialog()
    {
        setCloseOnEsc( true );
        setWidth( DIALOG_WIDTH );
        setHeight( DIALOG_HEIGHT );
    }

    public SelectServerDialog( List<ServerData> serverDatas, VirtualUtil virtualUtil )
    {
        this();

        // auto select server buttons
        Button autoSelectServerButton = new Button( virtualUtil.getTranslation( Translation.BUTTON_APPLICATIONS_SERVER_AUTO ), event -> {
            Collections.shuffle( this.serverDatas );
            setSelectedServers( this.serverDatas );
            close();
        } );
        autoSelectServerButton.setDisableOnClick( true );
        autoSelectServerButton.getElement()
                              .setAttribute( ATTRIBUTE_THEME, ATTRIBUTE_THEME_VALUE );

        // close buttons
        Button closeButton = new Button( virtualUtil.getTranslation( Translation.BUTTON_CLOSE ), event -> close() );
        closeButton.setDisableOnClick( true );
        closeButton.getElement()
                   .setAttribute( ATTRIBUTE_THEME, ATTRIBUTE_THEME_VALUE );

        // select server buttons
        Button selectServerButton = new Button( virtualUtil.getTranslation( Translation.BUTTON_APPLICATIONS_SERVER ), event -> {
            setSelectedServers( new ArrayList<>( serverGrid.getSelectedItems() ) );
            close();
        } );
        selectServerButton.setDisableOnClick( true );
        selectServerButton.setEnabled( false );
        selectServerButton.addClassName( ATTRIBUTE_CLASS_BUTTON_VALUE );
        selectServerButton.getElement()
                          .setAttribute( ATTRIBUTE_THEME, ATTRIBUTE_THEME_VALUE );
        selectServerButton.getStyle()
                          .set( STYLE_MARGIN_LEFT, STYLE_MARGIN_0PX_VALUE );

        // buttons layout
        HorizontalLayout buttonsLayout = new HorizontalLayout( selectServerButton, autoSelectServerButton, closeButton );
        buttonsLayout.setJustifyContentMode( JustifyContentMode.CENTER );
        buttonsLayout.setMargin( true );
        buttonsLayout.setPadding( true );
        buttonsLayout.setSpacing( true );
        buttonsLayout.setWidth( STYLE_WIDTH_100PERCENT_VALUE );
        buttonsLayout.getStyle()
                     .set( STYLE_MARGIN_BOTTOM, STYLE_MARGIN_0PX_VALUE )
                     .set( STYLE_MARGIN_LEFT, STYLE_MARGIN_0PX_VALUE )
                     .set( STYLE_MARGIN_RIGHT, STYLE_MARGIN_0PX_VALUE )
                     .set( STYLE_PADDING_BOTTOM, STYLE_PADDING_VALUE )
                     .set( STYLE_PADDING_LEFT, STYLE_PADDING_VALUE )
                     .set( STYLE_PADDING_RIGHT, STYLE_PADDING_VALUE );

        // servers
        this.serverDatas = new ArrayList<>( serverDatas );
        dataProvider = new ListDataProvider<>( this.serverDatas );
        serverGrid = new Grid<>();
        serverGrid.setDataProvider( dataProvider );
        serverGrid.setPageSize( GRID_PAGE_SIZE );
        serverGrid.setSelectionMode( SelectionMode.SINGLE );
        serverGrid.getStyle()
                  .set( STYLE_DISPLAY, STYLE_DISPLAY_VALUE )
                  .set( STYLE_MARGIN_LEFT, STYLE_MARGIN_1PX_VALUE )
                  .set( STYLE_MARGIN_RIGHT, STYLE_MARGIN_1PX_VALUE );
        serverGrid.addItemDoubleClickListener( event -> {
            setSelectedServers( new ArrayList<>( Arrays.asList( event.getItem() ) ) );
            close();
        } );
        serverGrid.addSelectionListener( event -> selectServerButton.setEnabled( CollectionUtils.isNotEmpty( event.getAllSelectedItems() ) ) );

        // name
        Grid.Column<ServerData> nameColumn = serverGrid.addColumn( ServerData::getName )
                                                       .setFlexGrow( 1 )
                                                       .setWidth( STYLE_WIDTH_150PX_VALUE )
                                                       .setResizable( false )
                                                       .setHeader( virtualUtil.getTranslation( Translation.FIELD_NAME ) );

        // address
        Grid.Column<ServerData> addressColumn = serverGrid.addColumn( ServerData::getAddress )
                                                          .setFlexGrow( 1 )
                                                          .setWidth( STYLE_WIDTH_150PX_VALUE )
                                                          .setResizable( false )
                                                          .setHeader( virtualUtil.getTranslation( Translation.FIELD_ADDRESS ) );

        // filter: name
        HeaderRow filterRow = serverGrid.appendHeaderRow();
        TextField nameFilter = new TextField();
        nameFilter.setValueChangeMode( ValueChangeMode.EAGER );
        nameFilter.setPlaceholder( virtualUtil.getTranslation( Translation.FIELD_FILTER ) );
        nameFilter.setSizeFull();
        nameFilter.addValueChangeListener( event -> dataProvider.addFilter( serverData -> StringUtils.containsIgnoreCase( serverData.getName(), nameFilter.getValue() ) ) );
        filterRow.getCell( nameColumn )
                 .setComponent( nameFilter );

        // filter: address
        TextField addressFilter = new TextField();
        addressFilter.setValueChangeMode( ValueChangeMode.EAGER );
        addressFilter.setPlaceholder( virtualUtil.getTranslation( Translation.FIELD_FILTER ) );
        addressFilter.setSizeFull();
        addressFilter.addValueChangeListener( event -> dataProvider.addFilter( serverData -> StringUtils.containsIgnoreCase( serverData.getAddress(), addressFilter.getValue() ) ) );
        filterRow.getCell( addressColumn )
                 .setComponent( addressFilter );

        // server grid layout
        VerticalLayout serverGridLayout = new VerticalLayout( serverGrid );
        serverGridLayout.setMargin( false );
        serverGridLayout.setPadding( false );
        serverGridLayout.setSpacing( false );
        serverGridLayout.setSizeFull();

        // form layout
        formLayout = new FormLayout( serverGridLayout, buttonsLayout );
        formLayout.setClassName( DIALOG_CLASS );
        formLayout.setResponsiveSteps( new ResponsiveStep( FORM_MINIMAL_WIDTH, FORM_COLUMNS, LabelsPosition.TOP ) );
        formLayout.setWidth( STYLE_WIDTH_100PERCENT_VALUE );
        add( formLayout );
    }

    public abstract void setSelectedServers( List<ServerData> serverDatas );

    @Override
    protected void onAttach( AttachEvent attachEvent )
    {
        attachEvent.getUI()
                   .getPage()
                   .executeJs( Command.JS.CLIPBOARD_ACTIVE, false );
    }

    @Override
    protected void onDetach( DetachEvent detachEvent )
    {
        detachEvent.getUI()
                   .getPage()
                   .executeJs( Command.JS.DISABLE_BARS, false );
        detachEvent.getUI()
                   .getPage()
                   .executeJs( Command.JS.CLIPBOARD_ACTIVE, true );
    }

}
