/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.ui.servergroups;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.EnumMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dvirtual.database.Database;
import org.dussan.vaadin.dvirtual.database.columns.ServersGroupColumns;
import org.dussan.vaadin.dvirtual.database.data.ServerData;
import org.dussan.vaadin.dvirtual.database.data.ServersGroupData;
import org.dussan.vaadin.dvirtual.i18n.Translation;
import org.dussan.vaadin.dvirtual.runners.TooltipRunner;
import org.dussan.vaadin.dvirtual.utils.VirtualUtil;
import org.dussan.vaadin.dvirtual.utils.dialog.DialogInfoUtil;

import com.vaadin.flow.component.UI;
import com.vaadin.flow.component.button.Button;
import com.vaadin.flow.component.button.ButtonVariant;
import com.vaadin.flow.component.combobox.ComboBox;
import com.vaadin.flow.component.grid.Grid;
import com.vaadin.flow.component.grid.HeaderRow;
import com.vaadin.flow.component.html.Div;
import com.vaadin.flow.component.html.Label;
import com.vaadin.flow.component.icon.Icon;
import com.vaadin.flow.component.icon.VaadinIcon;
import com.vaadin.flow.component.orderedlayout.HorizontalLayout;
import com.vaadin.flow.component.orderedlayout.VerticalLayout;
import com.vaadin.flow.component.textfield.TextField;
import com.vaadin.flow.data.provider.DataProvider;
import com.vaadin.flow.data.provider.SortDirection;
import com.vaadin.flow.data.provider.SortOrder;
import com.vaadin.flow.data.renderer.ComponentRenderer;
import com.vaadin.flow.data.value.ValueChangeMode;

public class ServerGroupsUI
    extends VerticalLayout
{

    private static final long serialVersionUID = -5388407159083971045L;

    private static final int FILTER_VALUE_CHANGE_TIMEOUT = 500;

    private static final int MAX_ELEMENTS_IN_ROW = 4;

    private static final int VISIBLE_ROWS = 50;

    private static final long NO_ELEMENTS = 0;

    private static final String CLASS_SETTINGS = "settings";

    private static final String CLASS_GRID_GROUPS = "grid-groups";

    private static final String CLASS_GRID_GROUPS_BUTTON = "grid-groups-button";

    private static final String CLASS_GRID_GROUPS_ICON = "grid-groups-icon";

    private static final String CLASS_GRID_GROUPS_LABEL = "grid-groups-label";

    private static final String NAME_WITH_NOTE = "%s (%s)";

    private static final String STYLE_COLOR = "color";

    private static final String STYLE_COLOR_RED_VALUE = "red";

    private static final String STYLE_DISPLAY = "display";

    private static final String STYLE_DISPLAY_VALUE = "grid";

    private static final String STYLE_GRID_COMBOBOX = "grid-combobox";

    private static final String STYLE_HEIGHT = "height";

    private static final String STYLE_HEIGHT_16PX_VALUE = "16px";

    private static final String STYLE_MARGIN_BOTTOM = "margin-bottom";

    private static final String STYLE_MARGIN_LEFT = "margin-left";

    private static final String STYLE_MARGIN_RIGHT = "margin-right";

    private static final String STYLE_MARGIN_TOP = "margin-top";

    private static final String STYLE_MARGIN_0PX_VALUE = "0px";

    private static final String STYLE_MARGIN_1PX_VALUE = "1px";

    private static final String STYLE_TEXT_ALIGN = "text-align";

    private static final String STYLE_TEXT_ALIGN_VALUE = "center";

    private static final String STYLE_WIDTH = "width";

    private static final String STYLE_WIDTH_100PX_VALUE = "100px";

    private static final String STYLE_WIDTH_130PX_VALUE = "130px";

    private static final String STYLE_WIDTH_150PX_VALUE = "150px";

    private static final String STYLE_WIDTH_16PX_VALUE = "16px";

    private transient List<ServersGroupData> serversGroupDatas = null;

    private transient Map<ServersGroupColumns, String> filterMap = null;

    private Database database = null;

    private DataProvider<ServersGroupData, Void> dataProvider = null;

    private Grid<ServersGroupData> serversGroupGrid = null;

    private Grid.Column<ServersGroupData> editDeleteColumn = null;

    private VirtualUtil virtualUtil = null;

    public ServerGroupsUI()
    {
        setMargin( false );
        setSpacing( true );
        setSizeFull();
        addClassName( CLASS_SETTINGS );
        filterMap = new EnumMap<>( ServersGroupColumns.class );
    }

    public ServerGroupsUI( Database database, VirtualUtil virtualUtil )
    {
        this();
        this.database = database;
        this.virtualUtil = virtualUtil;

        dataProvider = DataProvider.fromCallbacks( query -> {
            int offset = query.getOffset();
            int limit = query.getLimit();

            List<String> orders = new ArrayList<>();
            for ( SortOrder<String> order : query.getSortOrders() )
            {
                orders.add( SortDirection.DESCENDING.equals( order.getDirection() ) ? Database.ORDER_DESCENDING + order.getSorted() : order.getSorted() );
            }

            return new ArrayList<>( this.database.getServerGroups( offset, limit, filterMap, orders ) ).stream();
        }, query -> database.getServerGroupsRows( filterMap )
                            .intValue() );

        serversGroupDatas = new ArrayList<>( database.getAllServerGroups() );
        serversGroupGrid = new Grid<>();
        serversGroupGrid.setDataProvider( dataProvider );
        serversGroupGrid.setPageSize( VISIBLE_ROWS );
        serversGroupGrid.getStyle()
                        .set( STYLE_DISPLAY, STYLE_DISPLAY_VALUE )
                        .set( STYLE_MARGIN_LEFT, STYLE_MARGIN_1PX_VALUE )
                        .set( STYLE_MARGIN_RIGHT, STYLE_MARGIN_1PX_VALUE );
        serversGroupGrid.addItemDoubleClickListener( event -> {
            serversGroupGrid.select( event.getItem() );
            serversGroupEditDialog( event.getItem() ).open();
        } );
        add( serversGroupGrid );

        // name
        Grid.Column<ServersGroupData> nameColumn = serversGroupGrid.addColumn( ServersGroupData::getName )
                                                                   .setFlexGrow( 1 )
                                                                   .setResizable( true )
                                                                   .setSortable( true )
                                                                   .setSortProperty( ServersGroupColumns.NAME.getName() )
                                                                   .setWidth( STYLE_WIDTH_150PX_VALUE )
                                                                   .setHeader( virtualUtil.getTranslation( ServersGroupColumns.NAME.getValueForTranslation() ) );
        // note
        Grid.Column<ServersGroupData> noteColumn = serversGroupGrid.addColumn( ServersGroupData::getNote )
                                                                   .setFlexGrow( 3 )
                                                                   .setResizable( true )
                                                                   .setSortable( true )
                                                                   .setSortProperty( ServersGroupColumns.NOTE.getName() )
                                                                   .setHeader( virtualUtil.getTranslation( ServersGroupColumns.NOTE.getValueForTranslation() ) );
        // server
        Grid.Column<ServersGroupData> serverColumn = serversGroupGrid.addColumn( new ComponentRenderer<>( Div::new, ( serversDiv, serversGroupData ) -> {
            List<ServerData> serverDatas = new ArrayList<>( serversGroupData.getServerDatas() ).stream()
                                                                                               .sorted( Comparator.comparing( ServerData::getName ) )
                                                                                               .collect( Collectors.toList() );
            IntStream.range( 0, Math.min( MAX_ELEMENTS_IN_ROW, serverDatas.size() ) )
                     .filter( index -> Objects.nonNull( serverDatas.get( index ) ) )
                     .forEach( index -> {
                         Div serverDiv = new Div();
                         ServerData serverData = serverDatas.get( index );
                         if ( index < ( MAX_ELEMENTS_IN_ROW - 1 ) || serverDatas.size() <= MAX_ELEMENTS_IN_ROW )
                         {
                             Label label =
                                 new Label( StringUtils.isBlank( serverData.getNote() ) ? serverData.getName() : String.format( NAME_WITH_NOTE, serverData.getName(), serverData.getNote() ) );
                             label.addClassName( CLASS_GRID_GROUPS_LABEL );
                             Icon icon = serverData.isEnabled() ? new Icon( VaadinIcon.CHECK ) : new Icon( VaadinIcon.CLOSE );
                             icon.addClassName( CLASS_GRID_GROUPS_ICON );

                             serverDiv.addClassName( CLASS_GRID_GROUPS );
                             serverDiv.add( icon );
                             serverDiv.add( label );
                         }
                         else
                         {
                             Button serversButton = new Button( virtualUtil.getTranslation( Translation.BUTTON_SHOW ), event -> {
                                 serversGroupGrid.select( serversGroupData );
                                 serversGroupEditDialog( serversGroupData ).open();
                             } );
                             serversButton.setWidthFull();
                             serversButton.addClassName( CLASS_GRID_GROUPS_BUTTON );
                             serversButton.addThemeVariants( ButtonVariant.LUMO_CONTRAST );
                             serverDiv.addClassName( CLASS_GRID_GROUPS );
                             serverDiv.add( serversButton );
                             serverDiv.getStyle()
                                      .set( STYLE_MARGIN_BOTTOM, STYLE_MARGIN_0PX_VALUE )
                                      .set( STYLE_MARGIN_TOP, STYLE_MARGIN_0PX_VALUE );
                         }
                         serversDiv.add( serverDiv );
                     } );
        } ) )
                                                                     .setFlexGrow( 2 )
                                                                     .setResizable( true )
                                                                     .setHeader( virtualUtil.getTranslation( ServersGroupColumns.SERVER.getValueForTranslation() ) );
        // enabled
        Grid.Column<ServersGroupData> enabledColumn = serversGroupGrid.addColumn( new ComponentRenderer<>( Div::new, ( div, serversGroupData ) -> {
            Icon icon = serversGroupData.isEnabled() ? new Icon( VaadinIcon.CHECK ) : new Icon( VaadinIcon.CLOSE );
            icon.getStyle()
                .set( STYLE_HEIGHT, STYLE_HEIGHT_16PX_VALUE )
                .set( STYLE_WIDTH, STYLE_WIDTH_16PX_VALUE );
            div.add( icon );
            div.getStyle()
               .set( STYLE_TEXT_ALIGN, STYLE_TEXT_ALIGN_VALUE );
        } ) )
                                                                      .setFlexGrow( 0 )
                                                                      .setResizable( true )
                                                                      .setSortable( true )
                                                                      .setSortProperty( ServersGroupColumns.ENABLED.getName() )
                                                                      .setWidth( STYLE_WIDTH_130PX_VALUE )
                                                                      .setHeader( virtualUtil.getTranslation( ServersGroupColumns.ENABLED.getValueForTranslation() ) );
        // edit/delete data
        editDeleteColumn = serversGroupGrid.addColumn( new ComponentRenderer<>( serversGroupData -> {
            // delete
            Button deleteButton = new Button( new Icon( VaadinIcon.TRASH ), event -> {
                boolean noElements = database.getApplicationServersGroupRows( serversGroupData ) == NO_ELEMENTS;
                String message = noElements ? virtualUtil.getTranslation( Translation.SERVERS_GROUP_REMOVE, serversGroupData.getName() )
                                : virtualUtil.getTranslation( Translation.SERVERS_GROUP_REMOVE_APPLICATION, serversGroupData.getName() );
                ServersGroupRemoveDialog dialog = new ServersGroupRemoveDialog( message, virtualUtil )
                {
                    private static final long serialVersionUID = -1288407159083971045L;

                    @Override
                    public void savingData()
                    {
                        removeServersGroup( serversGroupData );
                    }
                };
                dialog.open();
                serversGroupGrid.select( serversGroupData );
            } );
            deleteButton.getStyle()
                        .set( STYLE_COLOR, STYLE_COLOR_RED_VALUE );

            // edit
            Button editButton = new Button( new Icon( VaadinIcon.EDIT ), event -> {
                serversGroupGrid.select( serversGroupData );
                serversGroupEditDialog( serversGroupData ).open();
            } );

            if ( virtualUtil.isTooltipsVisible() )
            {
                TooltipRunner tooltipRunner1 = new TooltipRunner( deleteButton, virtualUtil.getTranslation( Translation.BUTTON_DELETE ), UI.getCurrent() );
                virtualUtil.startRunnerService( tooltipRunner1 );
                TooltipRunner tooltipRunner2 = new TooltipRunner( editButton, virtualUtil.getTranslation( Translation.BUTTON_EDIT ), UI.getCurrent() );
                virtualUtil.startRunnerService( tooltipRunner2 );
            }

            HorizontalLayout layout = new HorizontalLayout( editButton, deleteButton );
            layout.setMargin( false );
            layout.setPadding( false );
            layout.setSpacing( false );
            layout.setSizeUndefined();
            return layout;
        } ) )
                                           .setFlexGrow( 0 )
                                           .setResizable( false )
                                           .setWidth( STYLE_WIDTH_100PX_VALUE );
        editDeleteColumn.setVisible( CollectionUtils.isNotEmpty( serversGroupDatas ) );

        // filter: name
        HeaderRow filterRow = serversGroupGrid.appendHeaderRow();
        TextField nameFilter = new TextField();
        nameFilter.setPlaceholder( virtualUtil.getTranslation( Translation.FIELD_FILTER ) );
        nameFilter.setSizeFull();
        nameFilter.setValueChangeMode( ValueChangeMode.LAZY );
        nameFilter.setValueChangeTimeout( FILTER_VALUE_CHANGE_TIMEOUT );
        nameFilter.addValueChangeListener( event -> {
            if ( nameFilter.isEmpty() )
            {
                filterMap.remove( ServersGroupColumns.NAME );
            }
            else
            {
                filterMap.put( ServersGroupColumns.NAME, nameFilter.getValue() );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( nameColumn )
                 .setComponent( nameFilter );

        // filter: note
        TextField noteFilter = new TextField();
        noteFilter.setPlaceholder( virtualUtil.getTranslation( Translation.FIELD_FILTER ) );
        noteFilter.setSizeFull();
        noteFilter.setValueChangeMode( ValueChangeMode.LAZY );
        noteFilter.setValueChangeTimeout( FILTER_VALUE_CHANGE_TIMEOUT );
        noteFilter.addValueChangeListener( event -> {
            if ( noteFilter.isEmpty() )
            {
                filterMap.remove( ServersGroupColumns.NOTE );
            }
            else
            {
                filterMap.put( ServersGroupColumns.NOTE, noteFilter.getValue() );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( noteColumn )
                 .setComponent( noteFilter );

        // servers group: server
        TextField serverFilter = new TextField();
        serverFilter.setPlaceholder( virtualUtil.getTranslation( Translation.FIELD_FILTER ) );
        serverFilter.setSizeFull();
        serverFilter.setValueChangeMode( ValueChangeMode.LAZY );
        serverFilter.setValueChangeTimeout( FILTER_VALUE_CHANGE_TIMEOUT );
        serverFilter.addValueChangeListener( event -> {
            if ( serverFilter.isEmpty() )
            {
                filterMap.remove( ServersGroupColumns.SERVER );
            }
            else
            {
                filterMap.put( ServersGroupColumns.SERVER, serverFilter.getValue() );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( serverColumn )
                 .setComponent( serverFilter );

        // filter: enabled
        String filterAll = virtualUtil.getTranslation( Translation.DATA_FILTER_ALL );
        String filterNo = virtualUtil.getTranslation( Translation.DATA_FILTER_NO );
        String filterYes = virtualUtil.getTranslation( Translation.DATA_FILTER_YES );
        ComboBox<String> enabledFilter = new ComboBox<>();
        enabledFilter.setAllowCustomValue( false );
        enabledFilter.setClassName( STYLE_GRID_COMBOBOX );
        enabledFilter.setItems( filterAll, filterYes, filterNo );
        enabledFilter.setValue( filterAll );
        enabledFilter.setSizeFull();
        enabledFilter.addValueChangeListener( event -> {
            if ( StringUtils.equals( enabledFilter.getValue(), filterYes ) )
            {
                filterMap.put( ServersGroupColumns.ENABLED, Boolean.TRUE.toString() );
            }
            else if ( StringUtils.equals( enabledFilter.getValue(), filterNo ) )
            {
                filterMap.put( ServersGroupColumns.ENABLED, Boolean.FALSE.toString() );
            }
            else
            {
                filterMap.remove( ServersGroupColumns.ENABLED );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( enabledColumn )
                 .setComponent( enabledFilter );
    }

    private ServersGroupEditDialog serversGroupEditDialog( ServersGroupData serversGroupData )
    {
        return ( new ServersGroupEditDialog( serversGroupData.getId(), serversGroupData, database, virtualUtil )
        {
            private static final long serialVersionUID = 5388407159083971045L;

            @Override
            public void savingData( ServersGroupData data )
            {
                if ( Objects.nonNull( data ) )
                {
                    boolean closeDialog = true;
                    if ( Objects.nonNull( database.getServersGroup( data.getId() ) ) )
                    {
                        closeDialog = updateServersGroup( data );
                    }
                    else
                    {
                        closeDialog = addServersGroup( data );
                    }

                    if ( closeDialog )
                    {
                        close();
                    }
                }
                else
                {
                    new DialogInfoUtil( virtualUtil.getTranslation( Translation.DATA_ERROR_SAVE ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
                }
            }
        } );
    }

    public boolean addServersGroup( ServersGroupData serversGroupData )
    {
        if ( Objects.nonNull( serversGroupData ) )
        {
            serversGroupData.setId( database.addServersGroup( serversGroupData ) );
            if ( !Objects.equals( serversGroupData.getId(), ServersGroupData.UNDEFINED_SERVERS_GROUP ) )
            {
                serversGroupDatas.add( serversGroupData );
                dataProvider.refreshAll();
                serversGroupGrid.select( serversGroupData );
                editDeleteColumn.setVisible( CollectionUtils.isNotEmpty( serversGroupDatas ) );
                return true;
            }
            else
            {
                new DialogInfoUtil( virtualUtil.getTranslation( Translation.SERVERS_GROUP_ERROR_ADD ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
            }
        }
        else
        {
            new DialogInfoUtil( virtualUtil.getTranslation( Translation.DATA_ERROR_SAVE ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
        }
        return false;
    }

    public boolean addServersGroup( long id, String name, String note, boolean enabled, List<ServerData> serverDatas )
    {
        return addServersGroup( new ServersGroupData( id, name, note, enabled, serverDatas ) );
    }

    public boolean removeServersGroup( ServersGroupData serversGroupData )
    {
        if ( Objects.nonNull( serversGroupData ) )
        {
            if ( database.removeServersGroup( serversGroupData ) )
            {
                serversGroupGrid.deselectAll();
                serversGroupDatas.remove( serversGroupData );
                dataProvider.refreshAll();
                editDeleteColumn.setVisible( CollectionUtils.isNotEmpty( serversGroupDatas ) );
                return true;
            }
            else
            {
                new DialogInfoUtil( virtualUtil.getTranslation( Translation.SERVERS_GROUP_ERROR_REMOVE ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
            }
        }
        else
        {
            new DialogInfoUtil( virtualUtil.getTranslation( Translation.DATA_ERROR_SAVE ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
        }
        return false;
    }

    public boolean removeServersGroup( int id )
    {
        return removeServersGroup( database.getServersGroup( id ) );
    }

    public boolean updateServersGroup( ServersGroupData serversGroupData )
    {
        if ( Objects.nonNull( serversGroupData ) )
        {
            if ( database.updateServersGroup( serversGroupData ) )
            {
                serversGroupDatas.removeIf( data -> Objects.equals( data.getId(), serversGroupData.getId() ) );
                serversGroupDatas.add( serversGroupData );
                dataProvider.refreshAll();
                serversGroupGrid.select( serversGroupData );
                editDeleteColumn.setVisible( CollectionUtils.isNotEmpty( serversGroupDatas ) );
                return true;
            }
            else
            {
                new DialogInfoUtil( virtualUtil.getTranslation( Translation.DATA_ERROR_SAVE, serversGroupData.getName() ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
            }
        }
        else
        {
            new DialogInfoUtil( virtualUtil.getTranslation( Translation.DATA_ERROR_SAVE ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
        }
        return false;
    }

    public boolean updateServersGroup( int id, String name, String note, boolean enabled, List<ServerData> serverDatas )
    {
        return updateServersGroup( new ServersGroupData( id, name, note, enabled, serverDatas ) );
    }

    public void addNewServersGroup()
    {
        ServersGroupEditDialog dialog = new ServersGroupEditDialog( ServersGroupData.NEW_SERVERS_GROUP, new ServersGroupData(), database, virtualUtil )
        {
            private static final long serialVersionUID = -1460198394796534539L;

            @Override
            public void savingData( ServersGroupData serversGroupData )
            {
                if ( Objects.nonNull( serversGroupData ) )
                {
                    serversGroupData.setId( database.addServersGroup( serversGroupData ) );
                    if ( !Objects.equals( serversGroupData.getId(), ServersGroupData.UNDEFINED_SERVERS_GROUP ) )
                    {
                        serversGroupDatas.add( serversGroupData );
                        dataProvider.refreshAll();
                        serversGroupGrid.select( serversGroupData );
                        editDeleteColumn.setVisible( CollectionUtils.isNotEmpty( serversGroupDatas ) );
                        close();
                    }
                    else
                    {
                        new DialogInfoUtil( virtualUtil.getTranslation( Translation.SERVERS_GROUP_ERROR_EXISTS, serversGroupData.getName() ),
                                            virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
                    }
                }
                else
                {
                    new DialogInfoUtil( virtualUtil.getTranslation( Translation.DATA_ERROR_SAVE ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
                }
            }
        };
        dialog.open();
    }

}
