/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.ui.servers;

import java.util.Objects;

import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dvirtual.database.data.ServerData;
import org.dussan.vaadin.dvirtual.i18n.Translation;
import org.dussan.vaadin.dvirtual.utils.VirtualUtil;
import org.dussan.vaadin.dvirtual.utils.dialog.DialogEditUtil;
import org.dussan.vaadin.dvirtual.utils.dialog.DialogInfoUtil;

import com.vaadin.flow.component.button.Button;
import com.vaadin.flow.component.checkbox.Checkbox;
import com.vaadin.flow.component.formlayout.FormLayout;
import com.vaadin.flow.component.formlayout.FormLayout.ResponsiveStep;
import com.vaadin.flow.component.formlayout.FormLayout.ResponsiveStep.LabelsPosition;
import com.vaadin.flow.component.html.Label;
import com.vaadin.flow.component.orderedlayout.FlexComponent.Alignment;
import com.vaadin.flow.component.orderedlayout.FlexComponent.JustifyContentMode;
import com.vaadin.flow.component.orderedlayout.HorizontalLayout;
import com.vaadin.flow.component.orderedlayout.VerticalLayout;
import com.vaadin.flow.component.textfield.IntegerField;
import com.vaadin.flow.component.textfield.TextField;
import com.vaadin.flow.data.binder.Binder;
import com.vaadin.flow.data.validator.StringLengthValidator;
import com.vaadin.flow.data.value.ValueChangeMode;
import com.vaadin.flow.function.SerializablePredicate;

public abstract class ServerEditDialog
    extends DialogEditUtil
{

    private static final long serialVersionUID = -8470555752877541909L;

    private static final int FORM_COLUMNS = 1;

    private static final int MAX_PORT_VALUE = ServerData.DEFAULT_XPRA_MAX_PORT;

    private static final int MIN_PORT_VALUE = 1;

    private static final int MIN_USERS_VALUE = 1;

    private static final int VALUE_CHANGE_TIMEOUT = 500;

    private static final String ATTRIBUTE_CLASS_BUTTON_VALUE = "primary";

    private static final String ATTRIBUTE_THEME = "theme";

    private static final String ATTRIBUTE_THEME_VALUE = "dark";

    private static final String DIALOG_CLASS = "edit-dialog";

    private static final String FORM_MINIMAL_WIDTH = "0";

    private static final String STYLE_MARGIN_BOTTOM = "margin-bottom";

    private static final String STYLE_MARGIN_LEFT = "margin-left";

    private static final String STYLE_MARGIN_RIGHT = "margin-right";

    private static final String STYLE_MARGIN_VALUE = "0px";

    private static final String STYLE_PADDING_BOTTOM = "padding-bottom";

    private static final String STYLE_PADDING_LEFT = "padding-left";

    private static final String STYLE_PADDING_RIGHT = "padding-right";

    private static final String STYLE_PADDING_TOP = "padding-top";

    private static final String STYLE_PADDING_VALUE = "0px";

    private static final String STYLE_WIDTH_100PERCENT_VALUE = "100%";

    private static final String STYLE_WIDTH_100PX_VALUE = "100px";

    private static final String STYLE_WIDTH_130PX_VALUE = "130px";

    private Checkbox enabledCheckbox = null;

    private IntegerField maxUsersIntegerField = null;

    private IntegerField sshPortIntegerField = null;

    private IntegerField xpraMaxPortIntegerField = null;

    private IntegerField xpraMinPortIntegerField = null;

    private TextField addressTextField = null;

    private TextField nameTextField = null;

    private TextField noteTextField = null;

    private TextField xpraVersionTextField = null;

    public ServerEditDialog( long serverId, ServerData serverData, VirtualUtil virtualUtil )
    {
        // save/close/test buttons
        Button saveButton = new Button( virtualUtil.getTranslation( Translation.BUTTON_SAVE ) );
        saveButton.addClassName( ATTRIBUTE_CLASS_BUTTON_VALUE );
        saveButton.setEnabled( false );
        saveButton.getElement()
                  .setAttribute( ATTRIBUTE_THEME, ATTRIBUTE_THEME_VALUE );
        saveButton.getStyle()
                  .set( STYLE_MARGIN_LEFT, STYLE_MARGIN_VALUE );
        Button closeButton = new Button( virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) );
        closeButton.getElement()
                   .setAttribute( ATTRIBUTE_THEME, ATTRIBUTE_THEME_VALUE );
        Button testConnectionButton = new Button( virtualUtil.getTranslation( Translation.BUTTON_CONNECT_TEST ), event -> {
            ServerTestDialog serverTestDialog = new ServerTestDialog( addressTextField.getValue(), sshPortIntegerField.getValue(), virtualUtil );
            serverTestDialog.open();
        } );
        testConnectionButton.getElement()
                            .setAttribute( ATTRIBUTE_THEME, ATTRIBUTE_THEME_VALUE );

        // buttons layout
        HorizontalLayout buttonsLayout = new HorizontalLayout( saveButton, testConnectionButton, closeButton );
        buttonsLayout.setJustifyContentMode( JustifyContentMode.CENTER );
        buttonsLayout.setMargin( true );
        buttonsLayout.setPadding( true );
        buttonsLayout.setSpacing( true );
        buttonsLayout.setWidth( STYLE_WIDTH_100PERCENT_VALUE );
        buttonsLayout.getStyle()
                     .set( STYLE_MARGIN_BOTTOM, STYLE_MARGIN_VALUE )
                     .set( STYLE_MARGIN_LEFT, STYLE_MARGIN_VALUE )
                     .set( STYLE_MARGIN_RIGHT, STYLE_MARGIN_VALUE )
                     .set( STYLE_PADDING_BOTTOM, STYLE_PADDING_VALUE )
                     .set( STYLE_PADDING_LEFT, STYLE_PADDING_VALUE )
                     .set( STYLE_PADDING_RIGHT, STYLE_PADDING_VALUE );

        // name
        nameTextField = new TextField( virtualUtil.getTranslation( Translation.FIELD_NAME ), StringUtils.stripToEmpty( serverData.getName() ),
                                       event -> saveButton.setEnabled( isAllRequiredValuesNonEmpty() && !StringUtils.equals( event.getValue(), event.getOldValue() ) ) );
        nameTextField.setAutofocus( true );
        nameTextField.setRequiredIndicatorVisible( true );
        nameTextField.setValueChangeMode( ValueChangeMode.LAZY );
        nameTextField.setValueChangeTimeout( VALUE_CHANGE_TIMEOUT );
        nameTextField.getStyle()
                     .set( STYLE_PADDING_TOP, STYLE_PADDING_VALUE );

        // XPRA min port
        Integer minXpraPortValue = Objects.requireNonNullElse( serverData.getXpraMinPort(), ServerData.DEFAULT_XPRA_MIN_PORT );
        xpraMinPortIntegerField = new IntegerField( virtualUtil.getTranslation( Translation.FIELD_XPRA_MIN ), minXpraPortValue, event -> {
            testConnectionButton.setEnabled( isAllConnectionValuesNonEmpty() );
            saveButton.setEnabled( isAllRequiredValuesNonEmpty() && !Objects.equals( event.getValue(), event.getOldValue() ) );
        } );
        xpraMinPortIntegerField.setRequiredIndicatorVisible( true );
        xpraMinPortIntegerField.setValueChangeMode( ValueChangeMode.LAZY );
        xpraMinPortIntegerField.setValueChangeTimeout( VALUE_CHANGE_TIMEOUT );
        xpraMinPortIntegerField.setWidth( STYLE_WIDTH_100PX_VALUE );

        // XPRA max port
        Integer maxXpraPortValue = Objects.requireNonNullElse( serverData.getXpraMaxPort(), ServerData.DEFAULT_XPRA_MAX_PORT );
        xpraMaxPortIntegerField = new IntegerField( virtualUtil.getTranslation( Translation.FIELD_XPRA_MAX ), maxXpraPortValue, event -> {
            testConnectionButton.setEnabled( isAllConnectionValuesNonEmpty() );
            saveButton.setEnabled( isAllRequiredValuesNonEmpty() && !Objects.equals( event.getValue(), event.getOldValue() ) );
        } );
        xpraMaxPortIntegerField.setRequiredIndicatorVisible( true );
        xpraMaxPortIntegerField.setValueChangeMode( ValueChangeMode.LAZY );
        xpraMaxPortIntegerField.setValueChangeTimeout( VALUE_CHANGE_TIMEOUT );
        xpraMaxPortIntegerField.setWidth( STYLE_WIDTH_100PX_VALUE );

        // XPRA version
        xpraVersionTextField = new TextField( virtualUtil.getTranslation( Translation.FIELD_XPRA_VERSION ), StringUtils.stripToEmpty( serverData.getXpraVersion() ),
                                              event -> saveButton.setEnabled( isAllRequiredValuesNonEmpty() && !StringUtils.equals( event.getValue(), event.getOldValue() ) ) );
        xpraVersionTextField.setReadOnly( true );
        xpraVersionTextField.setWidth( STYLE_WIDTH_130PX_VALUE );

        // enabled
        enabledCheckbox = new Checkbox( virtualUtil.getTranslation( Translation.FIELD_ENABLED ), serverId == ServerData.NEW_SERVER ? ServerData.DEFAULT_ENABLED : serverData.isEnabled() );
        enabledCheckbox.addValueChangeListener( event -> saveButton.setEnabled( isAllRequiredValuesNonEmpty() && !Objects.equals( event.getValue(), event.getOldValue() ) ) );

        // name layout: name, XPRA min port, XPRA max port, XPRA version, enabled
        HorizontalLayout nameLayout = new HorizontalLayout( nameTextField, xpraMinPortIntegerField, xpraMaxPortIntegerField, xpraVersionTextField, enabledCheckbox );
        nameLayout.expand( nameTextField );
        nameLayout.setDefaultVerticalComponentAlignment( Alignment.BASELINE );
        nameLayout.setPadding( false );
        nameLayout.setSpacing( true );
        nameLayout.setWidth( STYLE_WIDTH_100PERCENT_VALUE );

        // address
        addressTextField = new TextField( virtualUtil.getTranslation( Translation.FIELD_ADDRESS ), StringUtils.stripToEmpty( serverData.getAddress() ), event -> {
            testConnectionButton.setEnabled( isAllConnectionValuesNonEmpty() );
            saveButton.setEnabled( isAllRequiredValuesNonEmpty() && !StringUtils.equals( event.getValue(), event.getOldValue() ) );
        } );
        addressTextField.setValueChangeMode( ValueChangeMode.LAZY );
        addressTextField.setValueChangeTimeout( VALUE_CHANGE_TIMEOUT );
        addressTextField.setRequiredIndicatorVisible( true );

        // maximum users
        Integer maxUsersValue = Objects.requireNonNullElse( serverData.getMaxUsers(), ServerData.DEFAULT_MAX_USERS );
        maxUsersIntegerField = new IntegerField( virtualUtil.getTranslation( Translation.FIELD_USERS ), maxUsersValue, event -> {
            testConnectionButton.setEnabled( isAllConnectionValuesNonEmpty() );
            saveButton.setEnabled( isAllRequiredValuesNonEmpty() && !Objects.equals( event.getValue(), event.getOldValue() ) );
        } );
        maxUsersIntegerField.setRequiredIndicatorVisible( true );
        maxUsersIntegerField.setValueChangeMode( ValueChangeMode.LAZY );
        maxUsersIntegerField.setValueChangeTimeout( VALUE_CHANGE_TIMEOUT );
        maxUsersIntegerField.setWidth( STYLE_WIDTH_100PX_VALUE );

        // SSH port
        Integer sshPortValue = Objects.requireNonNullElse( serverData.getSshPort(), ServerData.DEFAULT_SSH_PORT );
        sshPortIntegerField = new IntegerField( virtualUtil.getTranslation( Translation.FIELD_SSH ), sshPortValue, event -> {
            testConnectionButton.setEnabled( isAllConnectionValuesNonEmpty() );
            saveButton.setEnabled( isAllRequiredValuesNonEmpty() && !Objects.equals( event.getValue(), event.getOldValue() ) );
        } );
        sshPortIntegerField.setRequiredIndicatorVisible( true );
        sshPortIntegerField.setValueChangeMode( ValueChangeMode.LAZY );
        sshPortIntegerField.setValueChangeTimeout( VALUE_CHANGE_TIMEOUT );
        sshPortIntegerField.setWidth( STYLE_WIDTH_100PX_VALUE );

        // address, maximum users, SSH port
        HorizontalLayout addressPortLayout = new HorizontalLayout( addressTextField, maxUsersIntegerField, sshPortIntegerField );
        addressPortLayout.expand( addressTextField );
        addressPortLayout.setDefaultVerticalComponentAlignment( Alignment.CENTER );
        addressPortLayout.setSpacing( true );
        addressPortLayout.setWidth( STYLE_WIDTH_100PERCENT_VALUE );

        // note
        noteTextField = new TextField( virtualUtil.getTranslation( Translation.FIELD_NOTE ), StringUtils.stripToEmpty( serverData.getNote() ),
                                       event -> saveButton.setEnabled( isAllRequiredValuesNonEmpty() && !StringUtils.equals( event.getValue(), event.getOldValue() ) ) );
        noteTextField.setValueChangeMode( ValueChangeMode.LAZY );
        noteTextField.setValueChangeTimeout( VALUE_CHANGE_TIMEOUT );
        noteTextField.setWidth( STYLE_WIDTH_100PERCENT_VALUE );

        // legend
        Label legendXpraMinPortLabel = new Label( virtualUtil.getTranslation( Translation.SERVER_XPRA_MIN ) );
        Label legendXpraMaxPortLabel = new Label( virtualUtil.getTranslation( Translation.SERVER_XPRA_MAX ) );
        Label legendMaxUsersLabel = new Label( virtualUtil.getTranslation( Translation.SERVER_USERS ) );
        VerticalLayout legendLayout = new VerticalLayout( legendXpraMinPortLabel, legendXpraMaxPortLabel, legendMaxUsersLabel );
        legendLayout.setSpacing( false );
        legendLayout.setWidth( STYLE_WIDTH_100PERCENT_VALUE );
        legendLayout.getStyle()
                    .set( STYLE_PADDING_BOTTOM, STYLE_PADDING_VALUE )
                    .set( STYLE_PADDING_LEFT, STYLE_PADDING_VALUE )
                    .set( STYLE_PADDING_RIGHT, STYLE_PADDING_VALUE );

        SerializablePredicate<Integer> minUsersPredicate = minUsersValue -> maxUsersIntegerField.getValue() >= MIN_USERS_VALUE;
        SerializablePredicate<Integer> maxSshPortMaxPredicate = maxSshPortValue -> sshPortIntegerField.getValue() <= MAX_PORT_VALUE;
        SerializablePredicate<Integer> maxSshPortMinPredicate = maxSshPortValue -> sshPortIntegerField.getValue() >= MIN_PORT_VALUE;
        SerializablePredicate<Integer> xpraMaxPortMaxPredicate = xpraMaxPortValue -> xpraMaxPortIntegerField.getValue() <= MAX_PORT_VALUE;
        SerializablePredicate<Integer> xpraMaxPortMinPredicate = xpraMaxPortValue -> xpraMaxPortIntegerField.getValue() >= MIN_PORT_VALUE;
        SerializablePredicate<Integer> xpraMinPortMaxPredicate = xpraMinPortValue -> xpraMinPortIntegerField.getValue() <= MAX_PORT_VALUE;
        SerializablePredicate<Integer> xpraMinPortMinPredicate = xpraMinPortValue -> xpraMinPortIntegerField.getValue() >= MIN_PORT_VALUE;

        Binder<ServerData> serverDataBinder = new Binder<>();
        serverDataBinder.forField( nameTextField )
                        .withValidator( new StringLengthValidator( virtualUtil.getTranslation( Translation.DATA_VALUE ), 1, null ) )
                        .bind( ServerData::getName, ServerData::setName );
        serverDataBinder.forField( addressTextField )
                        .withValidator( new StringLengthValidator( virtualUtil.getTranslation( Translation.DATA_VALUE ), 1, null ) )
                        .bind( ServerData::getAddress, ServerData::setAddress );
        serverDataBinder.forField( maxUsersIntegerField )
                        .withValidator( minUsersPredicate, virtualUtil.getTranslation( Translation.DATA_VALUE_1 ) )
                        .bind( ServerData::getMaxUsers, ServerData::setMaxUsers );
        serverDataBinder.forField( sshPortIntegerField )
                        .withValidator( maxSshPortMinPredicate, virtualUtil.getTranslation( Translation.DATA_VALUE_1 ) )
                        .withValidator( maxSshPortMaxPredicate, virtualUtil.getTranslation( Translation.DATA_VALUE_65535 ) )
                        .bind( ServerData::getSshPort, ServerData::setSshPort );
        serverDataBinder.forField( xpraMinPortIntegerField )
                        .withValidator( xpraMinPortMinPredicate, virtualUtil.getTranslation( Translation.DATA_VALUE_1 ) )
                        .withValidator( xpraMinPortMaxPredicate, virtualUtil.getTranslation( Translation.DATA_VALUE_65535 ) )
                        .bind( ServerData::getXpraMinPort, ServerData::setXpraMinPort );
        serverDataBinder.forField( xpraMaxPortIntegerField )
                        .withValidator( xpraMaxPortMinPredicate, virtualUtil.getTranslation( Translation.DATA_VALUE_1 ) )
                        .withValidator( xpraMaxPortMaxPredicate, virtualUtil.getTranslation( Translation.DATA_VALUE_65535 ) )
                        .bind( ServerData::getXpraMaxPort, ServerData::setXpraMaxPort );
        serverDataBinder.bind( enabledCheckbox, ServerData::isEnabled, ServerData::setEnabled );
        serverDataBinder.bind( noteTextField, ServerData::getNote, ServerData::setNote );

        testConnectionButton.setEnabled( isAllConnectionValuesNonEmpty() );
        saveButton.addClickListener( event -> {
            ServerData data = new ServerData();
            if ( Objects.nonNull( data ) && serverDataBinder.writeBeanIfValid( data ) )
            {
                data.setId( serverId );
                data.setXpraVersion( xpraVersionTextField.getValue() );
                savingData( data );
            }
            else
            {
                new DialogInfoUtil( virtualUtil.getTranslation( Translation.DATA_ERROR_SAVE ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
            }
        } );
        closeButton.addClickListener( event -> close() );

        // form layout
        setFormLayout( new FormLayout( nameLayout, addressPortLayout, noteTextField, legendLayout, buttonsLayout ) );
        getFormLayout().setClassName( DIALOG_CLASS );
        getFormLayout().setResponsiveSteps( new ResponsiveStep( FORM_MINIMAL_WIDTH, FORM_COLUMNS, LabelsPosition.TOP ) );
        getFormLayout().setWidth( STYLE_WIDTH_100PERCENT_VALUE );
        add( getFormLayout() );
    }

    public abstract void savingData( ServerData serversData );

    private boolean isAllConnectionValuesNonEmpty()
    {
        return ( StringUtils.isNotBlank( addressTextField.getValue() ) && Objects.nonNull( sshPortIntegerField.getValue() ) && sshPortIntegerField.getValue() >= MIN_PORT_VALUE );
    }

    private boolean isAllRequiredValuesNonEmpty()
    {
        return ( StringUtils.isNotBlank( addressTextField.getValue() ) && StringUtils.isNotBlank( nameTextField.getValue() ) && Objects.nonNull( xpraMinPortIntegerField.getValue() )
            && Objects.nonNull( xpraMaxPortIntegerField.getValue() ) && Objects.nonNull( maxUsersIntegerField.getValue() ) && Objects.nonNull( sshPortIntegerField.getValue() ) );
    }

}
