/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.ui.users;

import java.util.ArrayList;
import java.util.EnumMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.stream.Collectors;
import java.util.stream.LongStream;

import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dvirtual.database.Database;
import org.dussan.vaadin.dvirtual.database.columns.ApplicationColumns;
import org.dussan.vaadin.dvirtual.database.columns.ApplicationsGroupColumns;
import org.dussan.vaadin.dvirtual.database.data.ApplicationData;
import org.dussan.vaadin.dvirtual.database.data.ApplicationsGroupData;
import org.dussan.vaadin.dvirtual.database.data.UserData;
import org.dussan.vaadin.dvirtual.i18n.Translation;
import org.dussan.vaadin.dvirtual.utils.VirtualUtil;
import org.dussan.vaadin.dvirtual.utils.dialog.DialogEditUtil;
import org.dussan.vaadin.dvirtual.utils.dialog.DialogInfoUtil;

import com.vaadin.flow.component.button.Button;
import com.vaadin.flow.component.checkbox.Checkbox;
import com.vaadin.flow.component.combobox.ComboBox;
import com.vaadin.flow.component.formlayout.FormLayout;
import com.vaadin.flow.component.formlayout.FormLayout.ResponsiveStep;
import com.vaadin.flow.component.formlayout.FormLayout.ResponsiveStep.LabelsPosition;
import com.vaadin.flow.component.grid.Grid;
import com.vaadin.flow.component.grid.Grid.SelectionMode;
import com.vaadin.flow.component.grid.HeaderRow;
import com.vaadin.flow.component.html.Div;
import com.vaadin.flow.component.html.Label;
import com.vaadin.flow.component.icon.Icon;
import com.vaadin.flow.component.icon.VaadinIcon;
import com.vaadin.flow.component.orderedlayout.FlexComponent.Alignment;
import com.vaadin.flow.component.orderedlayout.FlexComponent.JustifyContentMode;
import com.vaadin.flow.component.orderedlayout.HorizontalLayout;
import com.vaadin.flow.component.orderedlayout.VerticalLayout;
import com.vaadin.flow.component.textfield.IntegerField;
import com.vaadin.flow.component.textfield.TextField;
import com.vaadin.flow.data.binder.Binder;
import com.vaadin.flow.data.provider.DataProvider;
import com.vaadin.flow.data.provider.SortDirection;
import com.vaadin.flow.data.provider.SortOrder;
import com.vaadin.flow.data.renderer.ComponentRenderer;
import com.vaadin.flow.data.validator.StringLengthValidator;
import com.vaadin.flow.data.value.ValueChangeMode;
import com.vaadin.flow.function.SerializablePredicate;

public abstract class UserEditDialog
    extends DialogEditUtil
{

    private static final long serialVersionUID = 9206161873749777193L;

    private static final int DEFAULT_MIN_VALUE = 1;

    private static final int FORM_COLUMNS = 1;

    private static final int NO_LIMITS = Integer.MAX_VALUE;

    private static final int TEXT_FIELD_MAX_LENGTH = 100;

    private static final int VALUE_CHANGE_TIMEOUT = 500;

    private static final int VISIBLE_ROWS = 50;

    private static final String ATTRIBUTE_CLASS_BUTTON_VALUE = "primary";

    private static final String ATTRIBUTE_THEME = "theme";

    private static final String ATTRIBUTE_THEME_VALUE = "dark";

    private static final String CLASS_DIALOG = "edit-dialog";

    private static final String CLASS_GRID_CELL_ACTIVE = "active";

    private static final String CLASS_GRID_CELL_NOT_ACTIVE = StringUtils.EMPTY;

    private static final String CLASS_GRID_CELL_THEME = "grid-cell-theme";

    private static final String CLASS_GRID_LABEL = "grid-label";

    private static final String FORM_MINIMAL_WIDTH = "0";

    private static final String STYLE_DISPLAY = "display";

    private static final String STYLE_DISPLAY_GRID_VALUE = "grid";

    private static final String STYLE_GRID_COMBOBOX = "grid-combobox";

    private static final String STYLE_HEIGHT = "height";

    private static final String STYLE_HEIGHT_16PX_VALUE = "16px";

    private static final String STYLE_MARGIN_BOTTOM = "margin-bottom";

    private static final String STYLE_MARGIN_LEFT = "margin-left";

    private static final String STYLE_MARGIN_RIGHT = "margin-right";

    private static final String STYLE_MARGIN_VALUE = "0px";

    private static final String STYLE_PADDING_14PX_VALUE = "14px";

    private static final String STYLE_PADDING_BOTTOM = "padding-bottom";

    private static final String STYLE_PADDING_LEFT = "padding-left";

    private static final String STYLE_PADDING_RIGHT = "padding-right";

    private static final String STYLE_PADDING_TOP = "padding-top";

    private static final String STYLE_PADDING_VALUE = "0px";

    private static final String STYLE_TEXT_ALIGN = "text-align";

    private static final String STYLE_TEXT_ALIGN_VALUE = "center";

    private static final String STYLE_WIDTH = "width";

    private static final String STYLE_WIDTH_100PERCENT_VALUE = "100%";

    private static final String STYLE_WIDTH_110PX_VALUE = "110px";

    private static final String STYLE_WIDTH_130PX_VALUE = "130px";

    private static final String STYLE_WIDTH_150PX_VALUE = "150px";

    private static final String STYLE_WIDTH_250PX_VALUE = "250px";

    private static final String STYLE_WIDTH_16PX_VALUE = "16px";

    private static final String STYLE_WIDTH_50PX_VALUE = "50px";

    private transient List<ApplicationData> selectedApplicationDatas = null;

    private transient List<ApplicationsGroupData> selectedApplicationsGroupDatas = null;

    private transient Map<ApplicationColumns, String> applicationFilterMap = null;

    private transient Map<ApplicationsGroupColumns, String> applicationsGroupFilterMap = null;

    private Checkbox administratorCheckbox = null;

    private Checkbox downloadCheckbox = null;

    private Checkbox enabledCheckbox = null;

    private Checkbox fullScreenCheckbox = null;

    private Checkbox screenshotCheckbox = null;

    private Checkbox uploadCheckbox = null;

    private Checkbox selectAllApplicationsCheckbox = null;

    private Checkbox selectAllUserGroupsCheckbox = null;

    private DataProvider<ApplicationData, Void> applicationDataProvider = null;

    private DataProvider<ApplicationsGroupData, Void> applicationsGroupDataProvider = null;

    private Grid<ApplicationData> applicationGrid = null;

    private Grid<ApplicationsGroupData> applicationsGroupGrid = null;

    private TextField fullNameTextField = null;

    private IntegerField maxApplicationsIntegerField = null;

    private IntegerField maxFileSizeIntegerField = null;

    private TextField userNameTextField = null;

    public UserEditDialog( long userId, UserData userData, Database database, VirtualUtil virtualUtil )
    {
        applicationFilterMap = new EnumMap<>( ApplicationColumns.class );
        applicationsGroupFilterMap = new EnumMap<>( ApplicationsGroupColumns.class );
        selectedApplicationDatas = new ArrayList<>();
        database.getUserApplications( userData )
                .forEach( userApplicaton -> selectedApplicationDatas.add( userApplicaton.getApplicationData() ) );
        selectedApplicationsGroupDatas = new ArrayList<>();
        database.getUserApplicationGroups( userData )
                .forEach( userApplicatonsGroup -> selectedApplicationsGroupDatas.add( userApplicatonsGroup.getApplicationsGroupData() ) );

        // save/close buttons
        Button saveButton = new Button( virtualUtil.getTranslation( Translation.BUTTON_SAVE ) );
        saveButton.setEnabled( false );
        saveButton.addClassName( ATTRIBUTE_CLASS_BUTTON_VALUE );
        saveButton.getElement()
                  .setAttribute( ATTRIBUTE_THEME, ATTRIBUTE_THEME_VALUE );
        saveButton.getStyle()
                  .set( STYLE_MARGIN_LEFT, STYLE_MARGIN_VALUE );
        Button closeButton = new Button( virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) );
        closeButton.getElement()
                   .setAttribute( ATTRIBUTE_THEME, ATTRIBUTE_THEME_VALUE );

        // buttons layout
        HorizontalLayout buttonsLayout = new HorizontalLayout( saveButton, closeButton );
        buttonsLayout.setJustifyContentMode( JustifyContentMode.CENTER );
        buttonsLayout.setMargin( true );
        buttonsLayout.setPadding( true );
        buttonsLayout.setSpacing( true );
        buttonsLayout.setWidth( STYLE_WIDTH_100PERCENT_VALUE );
        buttonsLayout.getStyle()
                     .set( STYLE_MARGIN_BOTTOM, STYLE_MARGIN_VALUE )
                     .set( STYLE_MARGIN_LEFT, STYLE_MARGIN_VALUE )
                     .set( STYLE_MARGIN_RIGHT, STYLE_MARGIN_VALUE )
                     .set( STYLE_PADDING_BOTTOM, STYLE_PADDING_VALUE )
                     .set( STYLE_PADDING_LEFT, STYLE_PADDING_VALUE )
                     .set( STYLE_PADDING_RIGHT, STYLE_PADDING_VALUE );

        // user name
        userNameTextField = new TextField( virtualUtil.getTranslation( Translation.FIELD_USERNAME ), StringUtils.stripToEmpty( userData.getUserName() ),
                                           event -> saveButton.setEnabled( isAllRequiredValuesNonEmpty() && !StringUtils.equals( event.getValue(), event.getOldValue() ) ) );
        userNameTextField.setAutofocus( true );
        userNameTextField.setMaxLength( TEXT_FIELD_MAX_LENGTH );
        userNameTextField.setRequiredIndicatorVisible( true );
        userNameTextField.setValueChangeMode( ValueChangeMode.LAZY );
        userNameTextField.setValueChangeTimeout( VALUE_CHANGE_TIMEOUT );
        userNameTextField.getStyle()
                         .set( STYLE_PADDING_TOP, STYLE_PADDING_VALUE );

        // administrator
        administratorCheckbox =
            new Checkbox( virtualUtil.getTranslation( Translation.FIELD_ADMINISTRATOR ), userId == UserData.NEW_USER ? UserData.DEFAULT_ADMINISTRATOR : userData.isAdministrator() );
        administratorCheckbox.addValueChangeListener( event -> saveButton.setEnabled( isAllRequiredValuesNonEmpty() && !Objects.equals( event.getValue(), event.getOldValue() ) ) );
        // enabled
        enabledCheckbox = new Checkbox( virtualUtil.getTranslation( Translation.FIELD_ENABLED ), userId == UserData.NEW_USER ? UserData.DEFAULT_ENABLED : userData.isEnabled() );
        enabledCheckbox.addValueChangeListener( event -> saveButton.setEnabled( isAllRequiredValuesNonEmpty() && !Objects.equals( event.getValue(), event.getOldValue() ) ) );

        // maximum applications
        int maxApplicationsValue = Objects.requireNonNullElse( userData.getMaxApplications(), UserData.DEFAULT_MAX_APPLICATIONS );
        maxApplicationsIntegerField = new IntegerField( virtualUtil.getTranslation( Translation.FIELD_APPLICATION_MAX ), maxApplicationsValue,
                                                        event -> saveButton.setEnabled( isAllRequiredValuesNonEmpty() && !Objects.equals( event.getValue(), event.getOldValue() ) ) );
        maxApplicationsIntegerField.setRequiredIndicatorVisible( true );
        maxApplicationsIntegerField.setValueChangeMode( ValueChangeMode.LAZY );
        maxApplicationsIntegerField.setValueChangeTimeout( VALUE_CHANGE_TIMEOUT );
        maxApplicationsIntegerField.setWidth( STYLE_WIDTH_110PX_VALUE );
        maxApplicationsIntegerField.getStyle()
                                   .set( STYLE_PADDING_TOP, STYLE_PADDING_VALUE );

        // maximum file size
        int maxFileSizeValue = Objects.requireNonNullElse( userData.getMaxFileSize(), UserData.DEFAULT_MAX_FILE_SIZE );
        maxFileSizeIntegerField = new IntegerField( virtualUtil.getTranslation( Translation.FIELD_FILE_SIZE ), maxFileSizeValue,
                                                    event -> saveButton.setEnabled( isAllRequiredValuesNonEmpty() && !Objects.equals( event.getValue(), event.getOldValue() ) ) );
        maxFileSizeIntegerField.setMin( DEFAULT_MIN_VALUE );
        maxFileSizeIntegerField.setRequiredIndicatorVisible( true );
        maxFileSizeIntegerField.setValueChangeMode( ValueChangeMode.LAZY );
        maxFileSizeIntegerField.setValueChangeTimeout( VALUE_CHANGE_TIMEOUT );
        maxFileSizeIntegerField.setWidth( STYLE_WIDTH_250PX_VALUE );
        maxFileSizeIntegerField.getStyle()
                               .set( STYLE_PADDING_TOP, STYLE_PADDING_VALUE );

        // user name layout: user name, maximum applications, maximum file size, administrator, enabled
        HorizontalLayout userNameLayout = new HorizontalLayout( userNameTextField, maxApplicationsIntegerField, maxFileSizeIntegerField, administratorCheckbox, enabledCheckbox );
        userNameLayout.expand( userNameTextField );
        userNameLayout.setDefaultVerticalComponentAlignment( Alignment.BASELINE );
        userNameLayout.setPadding( false );
        userNameLayout.setSpacing( true );
        userNameLayout.setWidth( STYLE_WIDTH_100PERCENT_VALUE );
        userNameLayout.getStyle()
                      .set( STYLE_PADDING_TOP, STYLE_PADDING_VALUE );

        // full name
        fullNameTextField = new TextField( virtualUtil.getTranslation( Translation.FIELD_FULLNAME ), StringUtils.stripToEmpty( userData.getFullName() ),
                                           event -> saveButton.setEnabled( isAllRequiredValuesNonEmpty() && !StringUtils.equals( event.getValue(), event.getOldValue() ) ) );
        fullNameTextField.setMaxLength( TEXT_FIELD_MAX_LENGTH );
        fullNameTextField.setValueChangeMode( ValueChangeMode.LAZY );
        fullNameTextField.setValueChangeTimeout( VALUE_CHANGE_TIMEOUT );
        fullNameTextField.setWidth( STYLE_WIDTH_100PERCENT_VALUE );

        // download
        downloadCheckbox = new Checkbox( virtualUtil.getTranslation( Translation.FIELD_DOWNLOAD ), userId == UserData.NEW_USER ? UserData.DEFAULT_DOWNLOAD : userData.isDownload() );
        downloadCheckbox.addValueChangeListener( event -> saveButton.setEnabled( isAllRequiredValuesNonEmpty() && !Objects.equals( event.getValue(), event.getOldValue() ) ) );
        // full screen
        fullScreenCheckbox = new Checkbox( virtualUtil.getTranslation( Translation.FIELD_FULL_SCREEN ), userId == UserData.NEW_USER ? UserData.DEFAULT_FULL_SCREEN : userData.isFullScreen() );
        fullScreenCheckbox.addValueChangeListener( event -> saveButton.setEnabled( isAllRequiredValuesNonEmpty() && !Objects.equals( event.getValue(), event.getOldValue() ) ) );
        // screenshot
        screenshotCheckbox = new Checkbox( virtualUtil.getTranslation( Translation.FIELD_SCREENSHOT ), userId == UserData.NEW_USER ? UserData.DEFAULT_SCREENSHOT : userData.isScreenshot() );
        screenshotCheckbox.addValueChangeListener( event -> saveButton.setEnabled( isAllRequiredValuesNonEmpty() && !Objects.equals( event.getValue(), event.getOldValue() ) ) );
        // upload
        uploadCheckbox = new Checkbox( virtualUtil.getTranslation( Translation.FIELD_UPLOAD ), userId == UserData.NEW_USER ? UserData.DEFAULT_UPLOAD : userData.isUpload() );
        uploadCheckbox.addValueChangeListener( event -> saveButton.setEnabled( isAllRequiredValuesNonEmpty() && !Objects.equals( event.getValue(), event.getOldValue() ) ) );

        // full name layout: full name, upload, download
        HorizontalLayout fullNameLayout = new HorizontalLayout( fullNameTextField, screenshotCheckbox, uploadCheckbox, downloadCheckbox, fullScreenCheckbox );
        fullNameLayout.expand( userNameTextField );
        fullNameLayout.setDefaultVerticalComponentAlignment( Alignment.BASELINE );
        fullNameLayout.setSpacing( true );
        fullNameLayout.setWidth( STYLE_WIDTH_100PERCENT_VALUE );
        fullNameLayout.getStyle()
                      .set( STYLE_PADDING_BOTTOM, STYLE_PADDING_14PX_VALUE );

        // application data provider
        applicationDataProvider = DataProvider.fromCallbacks( query -> {
            int offset = query.getOffset();
            int limit = query.getLimit();

            List<String> orders = new ArrayList<>();
            for ( SortOrder<String> order : query.getSortOrders() )
            {
                orders.add( order.getDirection()
                                 .equals( SortDirection.DESCENDING ) ? Database.ORDER_DESCENDING + order.getSorted() : order.getSorted() );
            }

            List<ApplicationData> datas = new ArrayList<>( database.getApplications( offset, limit, applicationFilterMap, orders ) );
            return datas.stream();
        }, query -> {
            long[] selectedApplicationIds = selectedApplicationDatas.stream()
                                                                    .mapToLong( ApplicationData::getId )
                                                                    .toArray();
            int filteredSelectedAplicationsSize = //
                database.getApplications( 0, NO_LIMITS, applicationFilterMap, null )
                        .stream()
                        .filter( applicationData -> LongStream.of( selectedApplicationIds )
                                                              .anyMatch( selectedApplicationId -> selectedApplicationId == applicationData.getId() ) )
                        .collect( Collectors.toList() )
                        .size();

            Long applicationRows = database.getApplicationRows( applicationFilterMap );
            selectAllApplicationsCheckbox.setValue( filteredSelectedAplicationsSize == applicationRows && applicationRows > 0 );

            return applicationRows.intValue();
        } );

        // application
        applicationGrid = new Grid<>();
        applicationGrid.setDataProvider( applicationDataProvider );
        applicationGrid.setId( CLASS_GRID_CELL_THEME );
        applicationGrid.setPageSize( VISIBLE_ROWS );
        applicationGrid.setSelectionMode( SelectionMode.NONE );
        applicationGrid.setWidth( STYLE_WIDTH_100PERCENT_VALUE );
        applicationGrid.getStyle()
                       .set( STYLE_DISPLAY, STYLE_DISPLAY_GRID_VALUE )
                       .set( STYLE_WIDTH, STYLE_WIDTH_100PERCENT_VALUE );

        // application: selected
        selectAllApplicationsCheckbox = new Checkbox();
        selectAllApplicationsCheckbox.addClickListener( event -> {
            selectedApplicationDatas.clear();
            if ( Boolean.TRUE.equals( event.getSource()
                                           .getValue() ) )
            {
                selectedApplicationDatas.addAll( database.getApplications( 0, NO_LIMITS, applicationFilterMap, null ) );
            }
            saveButton.setEnabled( isAllRequiredValuesNonEmpty() );
            applicationDataProvider.refreshAll();
        } );
        applicationGrid.addColumn( new ComponentRenderer<>( applicationData -> {
            Checkbox checkbox = new Checkbox();
            checkbox.setValue( selectedApplicationDatas.stream()
                                                       .anyMatch( selectedApplicationData -> Objects.equals( selectedApplicationData.getId(), applicationData.getId() ) ) );
            checkbox.addClickListener( event -> {
                if ( Boolean.TRUE.equals( event.getSource()
                                               .getValue() ) )
                {
                    selectedApplicationDatas.add( applicationData );
                }
                else
                {
                    selectedApplicationDatas.removeIf( selectedApplicationData -> Objects.equals( selectedApplicationData.getId(), applicationData.getId() ) );
                }
                saveButton.setEnabled( isAllRequiredValuesNonEmpty() );
                applicationDataProvider.refreshAll();
            } );
            return checkbox;
        } ) )
                       .setFlexGrow( 0 )
                       .setFrozen( true )
                       .setKey( ApplicationColumns.SELECTED.getName() )
                       .setResizable( false )
                       .setSortable( false )
                       .setWidth( STYLE_WIDTH_50PX_VALUE )
                       .setHeader( selectAllApplicationsCheckbox );

        // application: name
        Grid.Column<ApplicationData> nameApplicationColumn = applicationGrid.addColumn( ApplicationData::getName )
                                                                            .setFlexGrow( 1 )
                                                                            .setFrozen( true )
                                                                            .setKey( ApplicationColumns.NAME.getName() )
                                                                            .setResizable( true )
                                                                            .setSortable( true )
                                                                            .setSortProperty( ApplicationColumns.NAME.getName() )
                                                                            .setWidth( STYLE_WIDTH_150PX_VALUE )
                                                                            .setHeader( virtualUtil.getTranslation( ApplicationColumns.NAME.getValueForTranslation() ) );
        // application: note
        Grid.Column<ApplicationData> noteApplicationColumn = applicationGrid.addColumn( ApplicationData::getNote )
                                                                            .setFlexGrow( 3 )
                                                                            .setKey( ApplicationColumns.NOTE.getName() )
                                                                            .setResizable( true )
                                                                            .setSortable( true )
                                                                            .setSortProperty( ApplicationColumns.NOTE.getName() )
                                                                            .setHeader( virtualUtil.getTranslation( ApplicationColumns.NOTE.getValueForTranslation() ) );
        // application: enabled
        Grid.Column<ApplicationData> enabledApplicationColumn = applicationGrid.addColumn( new ComponentRenderer<>( Div::new, ( div, applicationData ) -> {
            Icon icon = applicationData.isEnabled() ? new Icon( VaadinIcon.CHECK ) : new Icon( VaadinIcon.CLOSE );
            icon.getStyle()
                .set( STYLE_HEIGHT, STYLE_HEIGHT_16PX_VALUE )
                .set( STYLE_WIDTH, STYLE_WIDTH_16PX_VALUE );
            div.add( icon );
            div.getStyle()
               .set( STYLE_TEXT_ALIGN, STYLE_TEXT_ALIGN_VALUE );
        } ) )
                                                                               .setFlexGrow( 0 )
                                                                               .setKey( ApplicationColumns.ENABLED.getName() )
                                                                               .setResizable( true )
                                                                               .setSortable( true )
                                                                               .setSortProperty( ApplicationColumns.ENABLED.getName() )
                                                                               .setWidth( STYLE_WIDTH_130PX_VALUE )
                                                                               .setHeader( virtualUtil.getTranslation( ApplicationColumns.ENABLED.getValueForTranslation() ) );

        Label applicationGridLabel = new Label( virtualUtil.getTranslation( Translation.APPLICATIONS ) );
        applicationGridLabel.setFor( virtualUtil.getTranslation( Translation.APPLICATIONS ) );
        applicationGridLabel.addClassName( CLASS_GRID_LABEL );

        VerticalLayout applicationGridLayout = new VerticalLayout( applicationGridLabel, applicationGrid );
        applicationGridLayout.setMargin( false );
        applicationGridLayout.setPadding( false );
        applicationGridLayout.setSpacing( false );
        applicationGridLayout.setSizeUndefined();
        applicationGridLayout.setWidth( STYLE_WIDTH_100PERCENT_VALUE );

        // applications group data provider
        applicationsGroupDataProvider = DataProvider.fromCallbacks( query -> {
            int offset = query.getOffset();
            int limit = query.getLimit();

            List<String> orders = new ArrayList<>();
            for ( SortOrder<String> order : query.getSortOrders() )
            {
                orders.add( order.getDirection()
                                 .equals( SortDirection.DESCENDING ) ? Database.ORDER_DESCENDING + order.getSorted() : order.getSorted() );
            }

            return new ArrayList<>( database.getApplicationGroups( offset, limit, applicationsGroupFilterMap, orders ) ).stream();
        }, query -> {
            long[] selectedApplicationsGroupIds = selectedApplicationsGroupDatas.stream()
                                                                                .mapToLong( ApplicationsGroupData::getId )
                                                                                .toArray();
            int filteredSelectedUserGroupsSize = //
                database.getApplicationGroups( 0, NO_LIMITS, applicationsGroupFilterMap, null )
                        .stream()
                        .filter( applicationsGroupData -> LongStream.of( selectedApplicationsGroupIds )
                                                                    .anyMatch( selectedApplicationsGroupId -> selectedApplicationsGroupId == applicationsGroupData.getId() ) )
                        .collect( Collectors.toList() )
                        .size();

            Long applicationsGroupRows = database.getApplicationsGroupRows( applicationsGroupFilterMap );
            selectAllUserGroupsCheckbox.setValue( filteredSelectedUserGroupsSize == applicationsGroupRows && applicationsGroupRows > 0 );

            return applicationsGroupRows.intValue();
        } );

        // applications group
        applicationsGroupGrid = new Grid<>();
        applicationsGroupGrid.setDataProvider( applicationsGroupDataProvider );
        applicationsGroupGrid.setId( CLASS_GRID_CELL_THEME );
        applicationsGroupGrid.setPageSize( VISIBLE_ROWS );
        applicationsGroupGrid.setSelectionMode( SelectionMode.NONE );
        applicationsGroupGrid.setWidth( STYLE_WIDTH_100PERCENT_VALUE );
        applicationsGroupGrid.getStyle()
                             .set( STYLE_DISPLAY, STYLE_DISPLAY_GRID_VALUE )
                             .set( STYLE_WIDTH, STYLE_WIDTH_100PERCENT_VALUE );

        // applications group: selected
        selectAllUserGroupsCheckbox = new Checkbox();
        selectAllUserGroupsCheckbox.addClickListener( event -> {
            selectedApplicationsGroupDatas.clear();
            if ( Boolean.TRUE.equals( event.getSource()
                                           .getValue() ) )
            {
                selectedApplicationsGroupDatas.addAll( database.getApplicationGroups( 0, NO_LIMITS, applicationsGroupFilterMap, null ) );
            }
            saveButton.setEnabled( isAllRequiredValuesNonEmpty() );
            applicationsGroupDataProvider.refreshAll();
        } );
        applicationsGroupGrid.addColumn( new ComponentRenderer<>( applicationsGroupData -> {
            Checkbox checkbox = new Checkbox();
            checkbox.setValue( selectedApplicationsGroupDatas.stream()
                                                             .anyMatch( selectedApplicationsGroupData -> Objects.equals( selectedApplicationsGroupData.getId(), applicationsGroupData.getId() ) ) );
            checkbox.addClickListener( event -> {
                if ( Boolean.TRUE.equals( event.getSource()
                                               .getValue() ) )
                {
                    selectedApplicationsGroupDatas.add( applicationsGroupData );
                }
                else
                {
                    selectedApplicationsGroupDatas.removeIf( selectedApplicationsGroupData -> Objects.equals( selectedApplicationsGroupData.getId(), applicationsGroupData.getId() ) );
                }
                saveButton.setEnabled( isAllRequiredValuesNonEmpty() );
                applicationsGroupDataProvider.refreshAll();
            } );
            return checkbox;
        } ) )
                             .setFlexGrow( 0 )
                             .setFrozen( true )
                             .setKey( ApplicationsGroupColumns.SELECTED.getName() )
                             .setResizable( false )
                             .setSortable( false )
                             .setWidth( STYLE_WIDTH_50PX_VALUE )
                             .setHeader( selectAllUserGroupsCheckbox );

        // applications group: name
        Grid.Column<ApplicationsGroupData> nameApplicationsGroupColumn = applicationsGroupGrid.addColumn( ApplicationsGroupData::getName )
                                                                                              .setFlexGrow( 1 )
                                                                                              .setFrozen( true )
                                                                                              .setKey( ApplicationsGroupColumns.NAME.getName() )
                                                                                              .setResizable( true )
                                                                                              .setSortable( true )
                                                                                              .setSortProperty( ApplicationsGroupColumns.NAME.getName() )
                                                                                              .setWidth( STYLE_WIDTH_150PX_VALUE )
                                                                                              .setHeader( virtualUtil.getTranslation( ApplicationsGroupColumns.NAME.getValueForTranslation() ) );
        // applications group: note
        Grid.Column<ApplicationsGroupData> noteApplicationsGroupColumn = applicationsGroupGrid.addColumn( ApplicationsGroupData::getNote )
                                                                                              .setFlexGrow( 3 )
                                                                                              .setKey( ApplicationsGroupColumns.NOTE.getName() )
                                                                                              .setResizable( true )
                                                                                              .setSortable( true )
                                                                                              .setSortProperty( ApplicationsGroupColumns.NOTE.getName() )
                                                                                              .setHeader( virtualUtil.getTranslation( ApplicationsGroupColumns.NOTE.getValueForTranslation() ) );
        // applications group: enabled
        Grid.Column<ApplicationsGroupData> enabledApplicationsGroupColumn = applicationsGroupGrid.addColumn( new ComponentRenderer<>( Div::new, ( div, applicationsGroupData ) -> {
            Icon icon = applicationsGroupData.isEnabled() ? new Icon( VaadinIcon.CHECK ) : new Icon( VaadinIcon.CLOSE );
            icon.getStyle()
                .set( STYLE_HEIGHT, STYLE_HEIGHT_16PX_VALUE )
                .set( STYLE_WIDTH, STYLE_WIDTH_16PX_VALUE );
            div.add( icon );
            div.getStyle()
               .set( STYLE_TEXT_ALIGN, STYLE_TEXT_ALIGN_VALUE );
        } ) )
                                                                                                 .setFlexGrow( 0 )
                                                                                                 .setKey( ApplicationsGroupColumns.ENABLED.getName() )
                                                                                                 .setResizable( true )
                                                                                                 .setSortable( true )
                                                                                                 .setSortProperty( ApplicationsGroupColumns.ENABLED.getName() )
                                                                                                 .setWidth( STYLE_WIDTH_130PX_VALUE )
                                                                                                 .setHeader( virtualUtil.getTranslation( ApplicationsGroupColumns.ENABLED.getValueForTranslation() ) );

        Label applicationsGroupGridLabel = new Label( virtualUtil.getTranslation( Translation.APPLICATION_GROUPS ) );
        applicationsGroupGridLabel.setFor( virtualUtil.getTranslation( Translation.APPLICATION_GROUPS ) );
        applicationsGroupGridLabel.addClassName( CLASS_GRID_LABEL );

        VerticalLayout applicationsGroupGridLayout = new VerticalLayout( applicationsGroupGridLabel, applicationsGroupGrid );
        applicationsGroupGridLayout.setMargin( false );
        applicationsGroupGridLayout.setPadding( false );
        applicationsGroupGridLayout.setSpacing( false );
        applicationsGroupGridLayout.setSizeUndefined();
        applicationsGroupGridLayout.setWidth( STYLE_WIDTH_100PERCENT_VALUE );

        // servers group and applications layout: servers group layout, applications layout
        HorizontalLayout serversGroupApplicationsLayout = new HorizontalLayout( applicationsGroupGridLayout, applicationGridLayout );
        serversGroupApplicationsLayout.setDefaultVerticalComponentAlignment( Alignment.CENTER );
        serversGroupApplicationsLayout.setSpacing( true );
        serversGroupApplicationsLayout.setWidth( STYLE_WIDTH_100PERCENT_VALUE );

        SerializablePredicate<Integer> minApplicationsPredicate = minUsersValue -> maxApplicationsIntegerField.getValue() >= DEFAULT_MIN_VALUE;
        SerializablePredicate<Integer> minFileSizePredicate = minUsersValue -> maxFileSizeIntegerField.getValue() >= DEFAULT_MIN_VALUE;

        Binder<UserData> userDataBinder = new Binder<>();
        userDataBinder.forField( userNameTextField )
                      .withValidator( new StringLengthValidator( virtualUtil.getTranslation( Translation.DATA_VALUE ), 1, null ) )
                      .bind( UserData::getUserName, UserData::setUserName );
        userDataBinder.bind( fullNameTextField, UserData::getFullName, UserData::setFullName );
        userDataBinder.forField( maxApplicationsIntegerField )
                      .withValidator( minApplicationsPredicate, virtualUtil.getTranslation( Translation.DATA_VALUE_1 ) )
                      .bind( UserData::getMaxApplications, UserData::setMaxApplications );
        userDataBinder.forField( maxFileSizeIntegerField )
                      .withValidator( minFileSizePredicate, virtualUtil.getTranslation( Translation.DATA_VALUE_1 ) )
                      .bind( UserData::getMaxFileSize, UserData::setMaxFileSize );
        userDataBinder.bind( administratorCheckbox, UserData::isAdministrator, UserData::setAdministrator );
        userDataBinder.bind( downloadCheckbox, UserData::isDownload, UserData::setDownload );
        userDataBinder.bind( enabledCheckbox, UserData::isEnabled, UserData::setEnabled );
        userDataBinder.bind( fullScreenCheckbox, UserData::isFullScreen, UserData::setFullScreen );
        userDataBinder.bind( screenshotCheckbox, UserData::isScreenshot, UserData::setScreenshot );
        userDataBinder.bind( uploadCheckbox, UserData::isUpload, UserData::setUpload );

        closeButton.addClickListener( event -> close() );
        saveButton.addClickListener( event -> {
            UserData data = new UserData();
            if ( userDataBinder.writeBeanIfValid( data ) )
            {
                data.setId( userId );
                savingData( data, selectedApplicationsGroupDatas, selectedApplicationDatas );
            }
            else
            {
                new DialogInfoUtil( virtualUtil.getTranslation( Translation.DATA_ERROR_SAVE ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
            }
        } );

        // applications group filter: name
        HeaderRow applicationsGroupFilterRow = applicationsGroupGrid.appendHeaderRow();
        TextField nameApplicationsGroupFilter = new TextField();
        nameApplicationsGroupFilter.setPlaceholder( virtualUtil.getTranslation( Translation.FIELD_FILTER ) );
        nameApplicationsGroupFilter.setSizeFull();
        nameApplicationsGroupFilter.setValueChangeMode( ValueChangeMode.LAZY );
        nameApplicationsGroupFilter.setValueChangeTimeout( VALUE_CHANGE_TIMEOUT );
        nameApplicationsGroupFilter.addValueChangeListener( event -> {
            if ( nameApplicationsGroupFilter.isEmpty() )
            {
                applicationsGroupFilterMap.remove( ApplicationsGroupColumns.NAME );
            }
            else
            {
                applicationsGroupFilterMap.put( ApplicationsGroupColumns.NAME, nameApplicationsGroupFilter.getValue() );
            }
            applicationsGroupDataProvider.refreshAll();
        } );
        applicationsGroupFilterRow.getCell( nameApplicationsGroupColumn )
                                  .setComponent( nameApplicationsGroupFilter );

        // applications group filter: note
        TextField noteApplicationsGroupFilter = new TextField();
        noteApplicationsGroupFilter.setPlaceholder( virtualUtil.getTranslation( Translation.FIELD_FILTER ) );
        noteApplicationsGroupFilter.setSizeFull();
        noteApplicationsGroupFilter.setValueChangeMode( ValueChangeMode.LAZY );
        noteApplicationsGroupFilter.setValueChangeTimeout( VALUE_CHANGE_TIMEOUT );
        noteApplicationsGroupFilter.addValueChangeListener( event -> {
            if ( noteApplicationsGroupFilter.isEmpty() )
            {
                applicationsGroupFilterMap.remove( ApplicationsGroupColumns.NOTE );
            }
            else
            {
                applicationsGroupFilterMap.put( ApplicationsGroupColumns.NOTE, noteApplicationsGroupFilter.getValue() );
            }
            applicationsGroupDataProvider.refreshAll();
        } );
        applicationsGroupFilterRow.getCell( noteApplicationsGroupColumn )
                                  .setComponent( noteApplicationsGroupFilter );

        // applications group filter: enabled
        String filterAll = virtualUtil.getTranslation( Translation.DATA_FILTER_ALL );
        String filterNo = virtualUtil.getTranslation( Translation.DATA_FILTER_NO );
        String filterYes = virtualUtil.getTranslation( Translation.DATA_FILTER_YES );
        ComboBox<String> enabledApplicationsGroupFilter = new ComboBox<>();
        enabledApplicationsGroupFilter.setAllowCustomValue( false );
        enabledApplicationsGroupFilter.setClassName( STYLE_GRID_COMBOBOX );
        enabledApplicationsGroupFilter.setItems( filterAll, filterYes, filterNo );
        enabledApplicationsGroupFilter.setValue( filterAll );
        enabledApplicationsGroupFilter.setSizeFull();
        enabledApplicationsGroupFilter.addValueChangeListener( event -> {
            if ( StringUtils.equals( enabledApplicationsGroupFilter.getValue(), filterYes ) )
            {
                applicationsGroupFilterMap.put( ApplicationsGroupColumns.ENABLED, Boolean.TRUE.toString() );
            }
            else if ( StringUtils.equals( enabledApplicationsGroupFilter.getValue(), filterNo ) )
            {
                applicationsGroupFilterMap.put( ApplicationsGroupColumns.ENABLED, Boolean.FALSE.toString() );
            }
            else
            {
                applicationsGroupFilterMap.remove( ApplicationsGroupColumns.ENABLED );
            }
            applicationsGroupDataProvider.refreshAll();
        } );
        applicationsGroupFilterRow.getCell( enabledApplicationsGroupColumn )
                                  .setComponent( enabledApplicationsGroupFilter );

        // application filter: name
        HeaderRow applicationFilterRow = applicationGrid.appendHeaderRow();
        TextField nameApplicationFilter = new TextField();
        nameApplicationFilter.setPlaceholder( virtualUtil.getTranslation( Translation.FIELD_FILTER ) );
        nameApplicationFilter.setSizeFull();
        nameApplicationFilter.setValueChangeMode( ValueChangeMode.LAZY );
        nameApplicationFilter.setValueChangeTimeout( VALUE_CHANGE_TIMEOUT );
        nameApplicationFilter.addValueChangeListener( event -> {
            if ( nameApplicationFilter.isEmpty() )
            {
                applicationFilterMap.remove( ApplicationColumns.NAME );
            }
            else
            {
                applicationFilterMap.put( ApplicationColumns.NAME, nameApplicationFilter.getValue() );
            }
            applicationDataProvider.refreshAll();
        } );
        applicationFilterRow.getCell( nameApplicationColumn )
                            .setComponent( nameApplicationFilter );

        // application filter: note
        TextField noteApplicationFilter = new TextField();
        noteApplicationFilter.setPlaceholder( virtualUtil.getTranslation( Translation.FIELD_FILTER ) );
        noteApplicationFilter.setSizeFull();
        noteApplicationFilter.setValueChangeMode( ValueChangeMode.LAZY );
        noteApplicationFilter.setValueChangeTimeout( VALUE_CHANGE_TIMEOUT );
        noteApplicationFilter.addValueChangeListener( event -> {
            if ( noteApplicationFilter.isEmpty() )
            {
                applicationFilterMap.remove( ApplicationColumns.NOTE );
            }
            else
            {
                applicationFilterMap.put( ApplicationColumns.NOTE, noteApplicationFilter.getValue() );
            }
            applicationDataProvider.refreshAll();
        } );
        applicationFilterRow.getCell( noteApplicationColumn )
                            .setComponent( noteApplicationFilter );

        // application filter: enabled
        ComboBox<String> enabledApplicationFilter = new ComboBox<>();
        enabledApplicationFilter.setAllowCustomValue( false );
        enabledApplicationFilter.setClassName( STYLE_GRID_COMBOBOX );
        enabledApplicationFilter.setItems( filterAll, filterYes, filterNo );
        enabledApplicationFilter.setValue( filterAll );
        enabledApplicationFilter.setSizeFull();
        enabledApplicationFilter.addValueChangeListener( event -> {
            if ( StringUtils.equals( enabledApplicationFilter.getValue(), filterYes ) )
            {
                applicationFilterMap.put( ApplicationColumns.ENABLED, Boolean.TRUE.toString() );
            }
            else if ( StringUtils.equals( enabledApplicationFilter.getValue(), filterNo ) )
            {
                applicationFilterMap.put( ApplicationColumns.ENABLED, Boolean.FALSE.toString() );
            }
            else
            {
                applicationFilterMap.remove( ApplicationColumns.ENABLED );
            }
            applicationDataProvider.refreshAll();
        } );
        applicationFilterRow.getCell( enabledApplicationColumn )
                            .setComponent( enabledApplicationFilter );

        // form layout
        setFormLayout( new FormLayout( userNameLayout, fullNameLayout, serversGroupApplicationsLayout, buttonsLayout ) );
        getFormLayout().setClassName( CLASS_DIALOG );
        getFormLayout().setResponsiveSteps( new ResponsiveStep( FORM_MINIMAL_WIDTH, FORM_COLUMNS, LabelsPosition.TOP ) );
        getFormLayout().setWidth( STYLE_WIDTH_100PERCENT_VALUE );
        add( getFormLayout() );

        // set active/selected rows
        applicationGrid.getColumnByKey( ApplicationColumns.ENABLED.getName() )
                       .setClassNameGenerator( applicationData -> getApplicationCellClassName( applicationData.getId() ) );
        applicationGrid.getColumnByKey( ApplicationColumns.NAME.getName() )
                       .setClassNameGenerator( applicationData -> getApplicationCellClassName( applicationData.getId() ) );
        applicationGrid.getColumnByKey( ApplicationColumns.NOTE.getName() )
                       .setClassNameGenerator( applicationData -> getApplicationCellClassName( applicationData.getId() ) );
        applicationGrid.getColumnByKey( ApplicationColumns.SELECTED.getName() )
                       .setClassNameGenerator( applicationData -> getApplicationCellClassName( applicationData.getId() ) );
        applicationsGroupGrid.getColumnByKey( ApplicationsGroupColumns.ENABLED.getName() )
                             .setClassNameGenerator( applicationsGroupData -> getApplicationsGroupCellClassName( applicationsGroupData.getId() ) );
        applicationsGroupGrid.getColumnByKey( ApplicationsGroupColumns.NAME.getName() )
                             .setClassNameGenerator( applicationsGroupData -> getApplicationsGroupCellClassName( applicationsGroupData.getId() ) );
        applicationsGroupGrid.getColumnByKey( ApplicationsGroupColumns.NOTE.getName() )
                             .setClassNameGenerator( applicationsGroupData -> getApplicationsGroupCellClassName( applicationsGroupData.getId() ) );
        applicationsGroupGrid.getColumnByKey( ApplicationsGroupColumns.SELECTED.getName() )
                             .setClassNameGenerator( applicationsGroupData -> getApplicationsGroupCellClassName( applicationsGroupData.getId() ) );
    }

    public abstract void savingData( UserData userData, List<ApplicationsGroupData> applicationsGroupDatas, List<ApplicationData> applicationDatas );

    private boolean isAllRequiredValuesNonEmpty()
    {
        return ( StringUtils.isNotBlank( userNameTextField.getValue() ) && Objects.nonNull( maxApplicationsIntegerField.getValue() ) && Objects.nonNull( maxFileSizeIntegerField.getValue() )
            && ( CollectionUtils.isNotEmpty( selectedApplicationDatas ) || CollectionUtils.isNotEmpty( selectedApplicationsGroupDatas ) ) );
    }

    private String getApplicationCellClassName( Long applicationDataId )
    {
        if ( selectedApplicationDatas.stream()
                                     .anyMatch( selectedApplicationData -> Objects.equals( selectedApplicationData.getId(), applicationDataId ) ) )
        {
            return CLASS_GRID_CELL_ACTIVE;
        }
        return CLASS_GRID_CELL_NOT_ACTIVE;
    }

    private String getApplicationsGroupCellClassName( Long applicationsGroupDataId )
    {
        if ( selectedApplicationsGroupDatas.stream()
                                           .anyMatch( selectedApplicationsGroupData -> Objects.equals( selectedApplicationsGroupData.getId(), applicationsGroupDataId ) ) )
        {
            return CLASS_GRID_CELL_ACTIVE;
        }
        return CLASS_GRID_CELL_NOT_ACTIVE;
    }

}
