/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.dussan.vaadin.dvirtual.utils;

import java.io.Serializable;
import java.math.BigDecimal;
import java.sql.Timestamp;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Objects;

import org.apache.commons.lang3.StringUtils;

public class DataUtil
    implements Serializable
{

    private static final long serialVersionUID = -1008789573296273850L;

    private static final String DATE_FORMAT = "yyyy-MM-dd HH:mm:ss";

    private DataUtil()
    {
        // nothing to do
    }

    public static final BigDecimal getBigdecimal( Object value )
    {
        if ( Objects.nonNull( getString( value ) ) )
        {
            try
            {
                return new BigDecimal( getString( value ) );
            }
            catch ( Exception exception )
            {
                return null;
            }
        }
        return null;
    }

    public static final boolean getBoolean( Object value )
    {
        if ( Objects.nonNull( getString( value ) ) )
        {
            try
            {
                if ( value instanceof Boolean )
                {
                    return (boolean) value;
                }
                if ( value instanceof Integer )
                {
                    return ( getInteger( value ) == 1 );
                }
                else
                {
                    return Boolean.getBoolean( getString( value ) );
                }
            }
            catch ( Exception exception )
            {
                return false;
            }
        }
        return false;
    }

    public static final Date getDate( Object value )
    {
        if ( Objects.nonNull( getString( value ) ) )
        {
            SimpleDateFormat dateFormat = new SimpleDateFormat( DATE_FORMAT );
            try
            {
                return dateFormat.parse( getString( value ) );
            }
            catch ( ParseException exception )
            {
                return null;
            }
        }
        return null;
    }

    public static final Double getDouble( Object value )
    {
        if ( Objects.nonNull( getString( value ) ) )
        {
            try
            {
                Double duble = Double.parseDouble( getString( value ) );
                if ( Objects.nonNull( duble ) )
                {
                    return duble;
                }
            }
            catch ( NullPointerException | NumberFormatException exception )
            {
                return null;
            }
        }
        return null;
    }

    public static final Integer getInteger( Object value )
    {
        if ( Objects.nonNull( getString( value ) ) )
        {
            try
            {
                Double duble = getDouble( value );
                if ( Objects.nonNull( duble ) )
                {
                    return duble.intValue();
                }
            }
            catch ( NullPointerException | NumberFormatException exception )
            {
                return null;
            }
        }
        return null;
    }

    public static final Long getLong( Object value )
    {
        if ( Objects.nonNull( getString( value ) ) )
        {
            Double duble = Double.parseDouble( getString( value ) );
            if ( Objects.nonNull( duble ) )
            {
                return duble.longValue();
            }
        }
        return null;
    }

    public static final String getString( Object value )
    {
        if ( Objects.nonNull( value ) )
        {
            return StringUtils.strip( value.toString() );
        }
        return null;
    }

    public static final Timestamp getTimestamp( Object value )
    {
        if ( Objects.nonNull( getString( value ) ) )
        {
            Date date = getDate( value );
            if ( Objects.nonNull( date ) )
            {
                return new Timestamp( date.getTime() );
            }
        }
        return null;
    }

    public static final Timestamp getTimestamp( Long value )
    {
        if ( Objects.nonNull( value ) )
        {
            return new Timestamp( value );
        }
        return null;
    }

}
