/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.utils;

import org.apache.commons.lang3.StringUtils;

import com.vaadin.flow.component.button.Button;
import com.vaadin.flow.component.button.ButtonVariant;
import com.vaadin.flow.component.html.Label;
import com.vaadin.flow.component.notification.Notification;
import com.vaadin.flow.component.orderedlayout.FlexComponent.Alignment;
import com.vaadin.flow.component.orderedlayout.HorizontalLayout;
import com.vaadin.flow.component.orderedlayout.VerticalLayout;
import com.vaadin.flow.component.progressbar.ProgressBar;
import com.vaadin.flow.component.progressbar.ProgressBarVariant;

public class NotificationUtil
    extends Notification
{

    private static final long serialVersionUID = 793735807271512301L;

    public static final int NOTIFICATION_CLOSE_DELAY = 2000;

    public static final int NOTIFICATION_CLOSE_LONG_DELAY = 5000;

    public static final double PROGRESS_5PERCENT = 0.5;

    public static final double PROGRESS_10PERCENT = 1;

    public static final double PROGRESS_20PERCENT = 2;

    public static final double PROGRESS_30PERCENT = 3;

    public static final double PROGRESS_40PERCENT = 4;

    public static final double PROGRESS_50PERCENT = 5;

    public static final double PROGRESS_60PERCENT = 6;

    public static final double PROGRESS_70PERCENT = 7;

    public static final double PROGRESS_80PERCENT = 8;

    public static final double PROGRESS_90PERCENT = 9;

    public static final double PROGRESS_START = 0;

    public static final double PROGRESS_STOP = 10;

    private static final int STRING_MESSAGE_LENGTH = 100;

    private static final String STYLE_HEIGHT_VALUE = "13px";

    private static final String STYLE_MIN_WIDTH = "min-width";

    private static final String STYLE_MIN_WIDTH_VALUE = "50vw";

    private static final String STYLE_WHITE_SPACE = "white-space";

    private static final String STYLE_WHITE_SPACE_VALUE = "nowrap";

    private static final String STYLE_WIDTH_VALUE = "100%";

    private boolean stopped = false;

    private Button button = null;

    private Label messageLabel = null;

    private Label statusLabel = null;

    private ProgressBar progressBar = null;

    public NotificationUtil()
    {
        setPosition( Position.BOTTOM_START );

        // message
        messageLabel = new Label();
        messageLabel.getElement()
                    .getStyle()
                    .set( STYLE_WHITE_SPACE, STYLE_WHITE_SPACE_VALUE );
        HorizontalLayout messageLayout = new HorizontalLayout();
        messageLayout.setDefaultVerticalComponentAlignment( Alignment.CENTER );
        messageLayout.setSizeFull();
        messageLayout.add( messageLabel );

        // status
        statusLabel = new Label();
        statusLabel.setVisible( false );
        statusLabel.getElement()
                   .getStyle()
                   .set( STYLE_WHITE_SPACE, STYLE_WHITE_SPACE_VALUE );

        // progress bar
        progressBar = new ProgressBar( PROGRESS_START, PROGRESS_STOP, PROGRESS_START );
        progressBar.setHeight( STYLE_HEIGHT_VALUE );
        progressBar.setWidth( STYLE_WIDTH_VALUE );
        progressBar.setVisible( false );
        progressBar.addThemeVariants( ProgressBarVariant.LUMO_CONTRAST );

        // button
        button = new Button();
        button.setDisableOnClick( true );
        button.setEnabled( false );
        button.setVisible( false );
        button.setWidth( STYLE_WIDTH_VALUE );
        button.addThemeVariants( ButtonVariant.LUMO_PRIMARY );

        VerticalLayout notifictionLayout = new VerticalLayout();
        notifictionLayout.setMargin( false );
        notifictionLayout.setPadding( false );
        notifictionLayout.setSpacing( true );
        notifictionLayout.setSizeUndefined();
        notifictionLayout.setWidth( STYLE_WIDTH_VALUE );
        notifictionLayout.add( messageLayout, statusLabel, progressBar );

        HorizontalLayout mainLayout = new HorizontalLayout( notifictionLayout, button );
        mainLayout.setVerticalComponentAlignment( Alignment.CENTER, button );
        mainLayout.expand( notifictionLayout );
        add( mainLayout );
    }

    public NotificationUtil( String notification )
    {
        setText( notification );
        setPosition( Position.BOTTOM_START );
    }

    public boolean isStopped()
    {
        return stopped;
    }

    public boolean isNotStopped()
    {
        return !isStopped();
    }

    public void setStopped( boolean stopped )
    {
        this.stopped = stopped;
    }

    public String getMessage()
    {
        return messageLabel.getText();
    }

    public void setMessage( String message )
    {
        messageLabel.setText( message );
    }

    public String getStatus()
    {
        return statusLabel.getText();
    }

    public void setStatus( String status )
    {
        statusLabel.setText( status );
    }

    public void setStatusVisible( boolean visible )
    {
        statusLabel.setVisible( visible );
    }

    public boolean isProgressBarVisible()
    {
        return progressBar.isVisible();
    }

    public void setProgressBarVisible( boolean visible )
    {
        progressBar.setVisible( visible );
    }

    public double getProgressValue()
    {
        return progressBar.getValue();
    }

    public void setProgressValue( double value )
    {
        progressBar.setValue( value );
        button.setEnabled( isNotStopped() && button.isEnabled() );
        if ( value == PROGRESS_STOP )
        {
            button.setEnabled( true );
            if ( StringUtils.length( getMessage() ) > STRING_MESSAGE_LENGTH )
            {
                messageLabel.getElement()
                            .getStyle()
                            .set( STYLE_MIN_WIDTH, STYLE_MIN_WIDTH_VALUE )
                            .set( STYLE_WHITE_SPACE, StringUtils.EMPTY );
            }
        }
    }

    public Button getButton()
    {
        return button;
    }

}
