/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.utils.dialog;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.util.Objects;
import java.util.UUID;

import org.dussan.vaadin.dvirtual.i18n.Translation;
import org.dussan.vaadin.dvirtual.utils.VirtualUtil;
import org.dussan.vaadin.dvirtual.utils.LoggerUtil;
import org.dussan.vaadin.dvirtual.utils.dialog.common.DialogUtil;
import org.slf4j.Logger;

import com.vaadin.flow.component.DetachEvent;
import com.vaadin.flow.component.button.Button;
import com.vaadin.flow.component.combobox.ComboBox;
import com.vaadin.flow.component.html.Label;
import com.vaadin.flow.component.orderedlayout.HorizontalLayout;
import com.vaadin.flow.component.orderedlayout.VerticalLayout;
import com.vaadin.flow.component.orderedlayout.FlexComponent.Alignment;
import com.vaadin.flow.component.orderedlayout.FlexComponent.JustifyContentMode;
import com.vaadin.flow.component.upload.Upload;

public abstract class DataImportDialog
    extends DialogUtil
{

    private static final long serialVersionUID = 5237591075367660354L;

    private static final Logger LOGGER = LoggerUtil.getLogger();

    private static final int MAX_FILES = 1;

    private static final int MAX_FILE_SIZE = 10000000;

    private static final String ATTRIBUTE_CLASS_BUTTON_VALUE = "primary";

    private static final String ATTRIBUTE_THEME = "theme";

    private static final String ATTRIBUTE_THEME_VALUE = "dark";

    private static final String FILE_TYPE = "text/xml";

    private static final String STYLE_DISPLAY = "display";

    private static final String STYLE_DISPLAY_NONE_VALUE = "none";

    private static final String STYLE_MARGIN_BOTTOM = "margin-bottom";

    private static final String STYLE_MARGIN_LEFT = "margin-left";

    private static final String STYLE_MARGIN_RIGHT = "margin-right";

    private static final String STYLE_MARGIN_VALUE = "0px";

    private static final String STYLE_PADDING_BOTTOM = "padding-bottom";

    private static final String STYLE_PADDING_LEFT = "padding-left";

    private static final String STYLE_PADDING_RIGHT = "padding-right";

    private static final String STYLE_PADDING_TOP = "padding-top";

    private static final String STYLE_PADDING_VALUE = "0px";

    private static final String STYLE_WHITE_SPACE = "white-space";

    private static final String STYLE_WHITE_SPACE_VALUE = "nowrap";

    private static final String STYLE_WIDTH_100PERCENT_VALUE = "100%";

    private static final String STYLE_WIDTH_450PX_VALUE = "450px";

    private static final String TEMPORARY_FILE_EXTENSION = ".dv_";

    private static final String TEMPORARY_FOLDER = System.getProperty( "java.io.tmpdir" );

    private Button importSelectFileButton = null;

    private ComboBox<String> importDataComboBox = null;

    private Upload upload = null;

    private VirtualUtil virtualUtil = null;

    public DataImportDialog( VirtualUtil virtualUtil )
    {
        this.virtualUtil = virtualUtil;
        setSizeUndefined();
        setWidth( STYLE_WIDTH_450PX_VALUE );

        // close button
        Button closeButton = new Button( virtualUtil.getTranslation( Translation.BUTTON_CLOSE ), event -> close() );
        closeButton.setDisableOnClick( true );
        closeButton.getElement()
                   .setAttribute( ATTRIBUTE_THEME, ATTRIBUTE_THEME_VALUE );

        // upload file button
        Button uploadFileButton = new Button( virtualUtil.getTranslation( Translation.BUTTON_IMPORT ), event -> {
            addImportDataComponent();
            importSelectFileButton.clickInClient();
        } );
        uploadFileButton.addClassName( ATTRIBUTE_CLASS_BUTTON_VALUE );
        uploadFileButton.getElement()
                        .setAttribute( ATTRIBUTE_THEME, ATTRIBUTE_THEME_VALUE );
        uploadFileButton.getStyle()
                        .set( STYLE_MARGIN_LEFT, STYLE_MARGIN_VALUE );

        // import data label, combo box
        Label importDataLabel = new Label( virtualUtil.getTranslation( Translation.BAR_NAVIGATION_IMPORT_DATA_SET ) );
        importDataLabel.getStyle()
                       .set( STYLE_WHITE_SPACE, STYLE_WHITE_SPACE_VALUE );
        importDataComboBox = new ComboBox<>();
        importDataComboBox.setItems( virtualUtil.getTranslation( Translation.BAR_NAVIGATION_SERVERS ), virtualUtil.getTranslation( Translation.BAR_NAVIGATION_SERVER_GROUPS ),
                                     virtualUtil.getTranslation( Translation.BAR_NAVIGATION_APPLICATIONS ), virtualUtil.getTranslation( Translation.BAR_NAVIGATION_APPLICATION_GROUPS ),
                                     virtualUtil.getTranslation( Translation.BAR_NAVIGATION_USERS ) );
        importDataComboBox.setValue( virtualUtil.getTranslation( Translation.BAR_NAVIGATION_SERVERS ) );
        importDataComboBox.setWidthFull();

        // import data layout
        HorizontalLayout importDataLayout = new HorizontalLayout( importDataLabel, importDataComboBox );
        importDataLayout.setJustifyContentMode( JustifyContentMode.CENTER );
        importDataLayout.setDefaultVerticalComponentAlignment( Alignment.CENTER );
        importDataLayout.setMargin( true );
        importDataLayout.setPadding( true );
        importDataLayout.setSpacing( true );
        importDataLayout.setWidth( STYLE_WIDTH_100PERCENT_VALUE );
        importDataLayout.getStyle()
                        .set( STYLE_MARGIN_BOTTOM, STYLE_MARGIN_VALUE )
                        .set( STYLE_MARGIN_LEFT, STYLE_MARGIN_VALUE )
                        .set( STYLE_MARGIN_RIGHT, STYLE_MARGIN_VALUE )
                        .set( STYLE_PADDING_BOTTOM, STYLE_PADDING_VALUE )
                        .set( STYLE_PADDING_LEFT, STYLE_PADDING_VALUE )
                        .set( STYLE_PADDING_RIGHT, STYLE_PADDING_VALUE )
                        .set( STYLE_PADDING_TOP, STYLE_PADDING_VALUE );

        // buttons layout
        HorizontalLayout buttonsLayout = new HorizontalLayout( uploadFileButton, closeButton );
        buttonsLayout.setJustifyContentMode( JustifyContentMode.CENTER );
        buttonsLayout.setMargin( true );
        buttonsLayout.setPadding( true );
        buttonsLayout.setSpacing( true );
        buttonsLayout.setWidth( STYLE_WIDTH_100PERCENT_VALUE );
        buttonsLayout.getStyle()
                     .set( STYLE_MARGIN_BOTTOM, STYLE_MARGIN_VALUE )
                     .set( STYLE_MARGIN_LEFT, STYLE_MARGIN_VALUE )
                     .set( STYLE_MARGIN_RIGHT, STYLE_MARGIN_VALUE )
                     .set( STYLE_PADDING_BOTTOM, STYLE_PADDING_VALUE )
                     .set( STYLE_PADDING_LEFT, STYLE_PADDING_VALUE )
                     .set( STYLE_PADDING_RIGHT, STYLE_PADDING_VALUE )
                     .set( STYLE_PADDING_TOP, STYLE_PADDING_VALUE );

        VerticalLayout mainLayout = new VerticalLayout();
        mainLayout.setMargin( false );
        mainLayout.setPadding( false );
        mainLayout.setSpacing( true );
        mainLayout.setSizeUndefined();
        mainLayout.setWidth( STYLE_WIDTH_100PERCENT_VALUE );
        mainLayout.add( importDataLayout, buttonsLayout );
        add( mainLayout );
    }

    public abstract void prepareImportingData();

    public abstract void importingData( String folder, String fileName, String fileType );

    public abstract void importingDataError( String errorMessage );

    private void addImportDataComponent()
    {
        // only use for select the file for the import
        importSelectFileButton = new Button( virtualUtil.getTranslation( Translation.BUTTON_IMPORT ) );

        String tmpFileName = UUID.randomUUID() + TEMPORARY_FILE_EXTENSION;
        upload = new Upload( ( fileName, mimeType ) -> {
            File tmpFile = new File( TEMPORARY_FOLDER, tmpFileName );
            try
            {
                return new FileOutputStream( tmpFile );
            }
            catch ( FileNotFoundException exception )
            {
                LOGGER.warn( exception.getMessage() );
                return null;
            }
        } );
        upload.setDropAllowed( false );
        upload.setAcceptedFileTypes( FILE_TYPE );
        upload.setMaxFiles( MAX_FILES );
        upload.setMaxFileSize( MAX_FILE_SIZE );
        upload.setUploadButton( importSelectFileButton );
        upload.setSizeUndefined();
        upload.getStyle()
              .set( STYLE_DISPLAY, STYLE_DISPLAY_NONE_VALUE );
        upload.addFailedListener( event -> importingDataError( event.getReason()
                                                                    .getMessage() ) );
        upload.addFileRejectedListener( event -> importingDataError( event.getErrorMessage() ) );
        upload.addStartedListener( event -> prepareImportingData() );
        upload.addSucceededListener( event -> importingData( TEMPORARY_FOLDER, tmpFileName, importDataComboBox.getValue() ) );
        add( upload );
    }

    @Override
    protected void onDetach( DetachEvent detachEvent )
    {
        if ( Objects.nonNull( upload ) )
        {
            remove( upload );
            importSelectFileButton = null;
            upload = null;
        }
    }

}
