/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.utils.dialog;

import java.util.Objects;

import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dvirtual.utils.dialog.common.DialogUtil;

import com.vaadin.flow.component.button.Button;
import com.vaadin.flow.component.html.Label;
import com.vaadin.flow.component.icon.Icon;
import com.vaadin.flow.component.icon.VaadinIcon;
import com.vaadin.flow.component.orderedlayout.FlexComponent.Alignment;
import com.vaadin.flow.component.orderedlayout.FlexComponent.JustifyContentMode;
import com.vaadin.flow.component.orderedlayout.HorizontalLayout;
import com.vaadin.flow.component.orderedlayout.VerticalLayout;
import com.vaadin.flow.component.progressbar.ProgressBar;

public class DialogConnectingUtil
    extends DialogUtil
{

    private static final long serialVersionUID = 5950596715395804508L;

    public static final double PROGRESS_5PERCENT = 0.5;

    public static final double PROGRESS_10PERCENT = 1;

    public static final double PROGRESS_20PERCENT = 2;

    public static final double PROGRESS_30PERCENT = 3;

    public static final double PROGRESS_40PERCENT = 4;

    public static final double PROGRESS_50PERCENT = 5;

    public static final double PROGRESS_60PERCENT = 6;

    public static final double PROGRESS_70PERCENT = 7;

    public static final double PROGRESS_80PERCENT = 8;

    public static final double PROGRESS_90PERCENT = 9;

    public static final double PROGRESS_99PERCENT = 9.999;

    public static final double PROGRESS_START = 0;

    public static final double PROGRESS_STOP = 10;

    private static final String ATTRIBUTE_THEME = "theme";

    private static final String ATTRIBUTE_THEME_VALUE = "dark";

    private static final String DIALOG_HEIGHT = "auto";

    private static final String DIALOG_WIDTH = "auto";

    private static final String STYLE_COLOR = "color";

    private static final String STYLE_COLOR_VALUE = "rgb(42, 127, 239)";

    private static final String STYLE_FONT_SIZE = "font-size";

    private static final String STYLE_FONT_SIZE_VALUE = "16px";

    private static final String STYLE_ICON_SIZE = "42px";

    private static final String STYLE_MARGIN_BOTTOM = "margin-bottom";

    private static final String STYLE_MARGIN_LEFT = "margin-left";

    private static final String STYLE_MARGIN_RIGHT = "margin-right";

    private static final String STYLE_MARGIN_VALUE = "0px";

    private static final String STYLE_PADDING_BOTTOM = "padding-bottom";

    private static final String STYLE_PADDING_LEFT = "padding-left";

    private static final String STYLE_PADDING_RIGHT = "padding-right";

    private static final String STYLE_PADDING_VALUE = "0px";

    private static final String STYLE_WIDTH_100PERCENT_VALUE = "100%";

    private boolean showDialog = false;

    private HorizontalLayout buttonsLayout = null;

    private Label messageLabel = null;

    private Label statusLabel = null;

    private ProgressBar progressBar = null;

    protected DialogConnectingUtil()
    {
        setCloseOnEsc( true );
        setWidth( DIALOG_WIDTH );
        setHeight( DIALOG_HEIGHT );
    }

    public DialogConnectingUtil( String message, String closeButtonText )
    {
        this();
        Icon icon = new Icon( VaadinIcon.EXCLAMATION_CIRCLE_O );
        icon.setSize( STYLE_ICON_SIZE );
        icon.getElement()
            .getStyle()
            .set( STYLE_COLOR, STYLE_COLOR_VALUE );

        // message
        messageLabel = new Label( message );
        messageLabel.getElement()
                    .getStyle()
                    .set( STYLE_FONT_SIZE, STYLE_FONT_SIZE_VALUE );

        HorizontalLayout messageLayout = new HorizontalLayout();
        messageLayout.setDefaultVerticalComponentAlignment( Alignment.CENTER );
        messageLayout.setSizeFull();
        messageLayout.add( icon, messageLabel );

        // status
        statusLabel = new Label( StringUtils.EMPTY );
        statusLabel.setVisible( false );
        statusLabel.getElement()
                   .getStyle()
                   .set( STYLE_FONT_SIZE, STYLE_FONT_SIZE_VALUE );

        // progress bar
        progressBar = new ProgressBar( PROGRESS_START, PROGRESS_STOP, PROGRESS_START );
        progressBar.setWidth( STYLE_WIDTH_100PERCENT_VALUE );

        // close
        Button closeButton = new Button( closeButtonText, event -> close() );
        closeButton.setAutofocus( true );
        closeButton.getElement()
                   .setAttribute( ATTRIBUTE_THEME, ATTRIBUTE_THEME_VALUE );
        buttonsLayout = new HorizontalLayout( closeButton );
        buttonsLayout.setJustifyContentMode( JustifyContentMode.CENTER );
        buttonsLayout.setMargin( true );
        buttonsLayout.setPadding( true );
        buttonsLayout.setSpacing( true );
        buttonsLayout.setWidth( STYLE_WIDTH_100PERCENT_VALUE );
        buttonsLayout.getElement()
                     .getStyle()
                     .set( STYLE_MARGIN_BOTTOM, STYLE_MARGIN_VALUE )
                     .set( STYLE_MARGIN_LEFT, STYLE_MARGIN_VALUE )
                     .set( STYLE_MARGIN_RIGHT, STYLE_MARGIN_VALUE )
                     .set( STYLE_PADDING_BOTTOM, STYLE_PADDING_VALUE )
                     .set( STYLE_PADDING_LEFT, STYLE_PADDING_VALUE )
                     .set( STYLE_PADDING_RIGHT, STYLE_PADDING_VALUE );

        VerticalLayout mainLayout = new VerticalLayout();
        mainLayout.setMargin( false );
        mainLayout.setPadding( false );
        mainLayout.setSpacing( true );
        mainLayout.setSizeUndefined();
        mainLayout.setWidth( STYLE_WIDTH_100PERCENT_VALUE );
        mainLayout.add( messageLayout, statusLabel, progressBar, buttonsLayout );
        add( mainLayout );
    }

    public boolean haveToShowDialog()
    {
        return showDialog;
    }

    public void setShowDialog( boolean showDialog )
    {
        this.showDialog = showDialog;
    }

    public void setButtonsVisible( boolean visible )
    {
        setCloseOnEsc( visible );
        if ( Objects.nonNull( buttonsLayout ) )
        {
            buttonsLayout.setVisible( visible );
            setProgressBarVisible( !visible );
            setStatusVisible( false );
        }
    }

    public String getMessage()
    {
        return messageLabel.getText();
    }

    public void setMessage( String message )
    {
        messageLabel.setText( message );
    }

    public String getStatus()
    {
        return statusLabel.getText();
    }

    public void setStatus( String status )
    {
        statusLabel.setText( status );
    }

    public void setStatusVisible( boolean visible )
    {
        statusLabel.setVisible( visible );
    }

    public boolean isProgressBarVisible()
    {
        return progressBar.isVisible();
    }

    public void setProgressBarVisible( boolean visible )
    {
        progressBar.setVisible( visible );
    }

    public double getProgressValue()
    {
        return progressBar.getValue();
    }

    public void setProgressValue( double value )
    {
        progressBar.setValue( value );
    }

}
