/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.xpra.io.rencode;

import java.io.EOFException;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInput;
import java.io.PushbackInputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;

import org.apache.commons.lang3.StringUtils;

/**
 * Rencode type codes forked from rencode-java.
 * 
 * @see {@linkplain https://github.com/jksiezni/rencode4j/}
 */
public class RencodeInputStream
    extends PushbackInputStream
    implements ObjectInput
{

    private byte[] buffer = new byte[8];

    private Decoder<?>[] decoders = new Decoder[256];

    private interface Decoder<T>
    {
        T decode( int token )
            throws IOException;
    }

    public RencodeInputStream( InputStream inputStream )
    {
        this( inputStream, false );
    }

    public RencodeInputStream( InputStream inputStream, boolean decodeAsString )
    {
        super( inputStream );

        decoders[Rencode.CHAR_TRUE] = new Decoder<Boolean>()
        {
            @Override
            public Boolean decode( int token )
                throws IOException
            {
                return Boolean.TRUE;
            }
        };

        decoders[Rencode.CHAR_FALSE] = new Decoder<Boolean>()
        {
            @Override
            public Boolean decode( int token )
                throws IOException
            {
                return Boolean.FALSE;
            }
        };

        decoders[Rencode.CHAR_NONE] = new Decoder<Object>()
        {
            @Override
            public Object decode( int token )
                throws IOException
            {
                return null;
            }
        };

        decoders['0'] = decoders['1'] = decoders['2'] = decoders['3'] = decoders['4'] = decoders['5'] = decoders['6'] = decoders['7'] = decoders['8'] = decoders['9'] = new Decoder<Object>()
        {
            @Override
            public Object decode( int token )
                throws IOException
            {
                final int length = fetchLength( token );
                if ( decodeAsString )
                {
                    return fetchString( length );
                }
                else
                {
                    final byte[] bytes = new byte[length];
                    readFully( bytes );

                    return bytes;
                }
            }
        };

        for ( int i = Rencode.STRING_FIXED_START; i < Rencode.STRING_FIXED_START + Rencode.STRING_FIXED_COUNT; ++i )
        {
            decoders[i] = new Decoder<Object>()
            {
                @Override
                public Object decode( int token )
                    throws IOException
                {
                    final int length = token - Rencode.STRING_FIXED_START;
                    if ( decodeAsString )
                    {
                        return fetchString( length );
                    }
                    else
                    {
                        final byte[] bytes = new byte[length];
                        readFully( bytes );

                        return bytes;
                    }
                }
            };
        }

        for ( int i = Rencode.INT_POS_FIXED_START; i < Rencode.INT_POS_FIXED_COUNT; ++i )
        {
            decoders[i] = new Decoder<Byte>()
            {
                @Override
                public Byte decode( int token )
                    throws IOException
                {
                    return (byte) token;
                }
            };
        }

        for ( int i = Rencode.INT_NEG_FIXED_START; i < Rencode.INT_NEG_FIXED_START + Rencode.INT_NEG_FIXED_COUNT; ++i )
        {
            decoders[i] = new Decoder<Byte>()
            {
                @Override
                public Byte decode( int token )
                    throws IOException
                {
                    return (byte) ( Rencode.INT_NEG_FIXED_START - 1 - token );
                }
            };
        }

        decoders[Rencode.CHAR_INT1] = new Decoder<Byte>()
        {
            @Override
            public Byte decode( int token )
                throws IOException
            {
                return (byte) read();
            }
        };

        decoders[Rencode.CHAR_INT2] = new Decoder<Short>()
        {
            @Override
            public Short decode( int token )
                throws IOException
            {
                return (short) ( read() << 8 | read() );
            }
        };

        decoders[Rencode.CHAR_INT4] = new Decoder<Integer>()
        {
            @Override
            public Integer decode( int token )
                throws IOException
            {
                return fetchInt();
            }
        };

        decoders[Rencode.CHAR_INT8] = new Decoder<Long>()
        {
            @Override
            public Long decode( int token )
                throws IOException
            {
                return fetchLong();
            }
        };

        decoders[Rencode.CHAR_FLOAT32] = new Decoder<Float>()
        {
            @Override
            public Float decode( int token )
                throws IOException
            {
                return Float.intBitsToFloat( fetchInt() );
            }
        };

        decoders[Rencode.CHAR_FLOAT64] = new Decoder<Double>()
        {
            @Override
            public Double decode( int token )
                throws IOException
            {
                return Double.longBitsToDouble( fetchLong() );
            }
        };

        for ( int i = Rencode.LIST_FIXED_START; i < ( Rencode.LIST_FIXED_START + Rencode.LIST_FIXED_COUNT ); ++i )
        {
            decoders[i] = new Decoder<List<?>>()
            {
                @Override
                public List<?> decode( int token )
                    throws IOException
                {
                    int count = token - Rencode.LIST_FIXED_START;
                    final List<Object> list = new ArrayList<>( count );
                    while ( count-- > 0 )
                    {
                        list.add( readObject() );
                    }

                    return list;
                }
            };
        }

        decoders[Rencode.CHAR_LIST] = new Decoder<List<?>>()
        {
            @Override
            public List<?> decode( int token )
                throws IOException
            {
                final List<Object> list = new ArrayList<>( 2 * Rencode.LIST_FIXED_COUNT );
                do
                {
                    list.add( readObject() );
                }
                while ( peek() != Rencode.CHAR_TERMINATION );
                read();

                return list;
            }
        };

        for ( int i = Rencode.DICTIONARY_FIXED_START; i < ( Rencode.DICTIONARY_FIXED_START + Rencode.DICTIONARY_FIXED_COUNT ); ++i )
        {
            decoders[i] = new Decoder<Map<?, ?>>()
            {
                @Override
                public Map<?, ?> decode( int token )
                    throws IOException
                {
                    int count = token - Rencode.DICTIONARY_FIXED_START;
                    final Map<Object, Object> map = new HashMap<>( count, 1 );
                    while ( count-- > 0 )
                    {
                        map.put( readKey(), readObject() );
                    }

                    return map;
                }
            };
        }

        decoders[Rencode.CHAR_DICT] = new Decoder<Map<?, ?>>()
        {
            @Override
            public Map<?, ?> decode( int token )
                throws IOException
            {
                final Map<Object, Object> map = new HashMap<>( 2 * Rencode.DICTIONARY_FIXED_COUNT );
                do
                {
                    map.put( readKey(), readObject() );
                }
                while ( peek() != Rencode.CHAR_TERMINATION );
                read();

                return map;
            }
        };
    }

    @Override
    public int read()
        throws IOException
    {
        final int value = super.read();
        if ( value < 0 )
        {
            throw new EOFException( "end of stream" );
        }

        return value;
    }

    @Override
    public Object readObject()
        throws IOException
    {
        final int token = read();
        if ( Objects.nonNull( decoders ) && Objects.nonNull( decoders[token] ) )
        {
            return decoders[token].decode( token );
        }
        return new Object();
    }

    @Override
    public void readFully( byte[] b )
        throws IOException
    {
        readFully( b, 0, b.length );
    }

    @Override
    public void readFully( byte[] b, int off, int len )
        throws IOException
    {
        int total = 0;
        while ( total < len )
        {
            final int r = read( b, off + total, len - total );
            if ( r < 0 )
            {
                throw new EOFException();
            }
            total += r;
        }
    }

    @Override
    public int skipBytes( int n )
        throws IOException
    {
        return (int) skip( n );
    }

    @Override
    public boolean readBoolean()
        throws IOException
    {
        return read() == Rencode.CHAR_TRUE;
    }

    private long readNumber()
        throws IOException
    {
        final int token = read();
        if ( 0 <= token && token < Rencode.INT_POS_FIXED_COUNT )
        {
            return token;
        }
        else if ( Rencode.INT_NEG_FIXED_START <= token && token < ( Rencode.INT_NEG_FIXED_START + Rencode.INT_NEG_FIXED_COUNT ) )
        {
            return Rencode.INT_NEG_FIXED_START - 1L - token;
        }
        switch ( token )
        {
            case Rencode.CHAR_INT1:
                return read();
            case Rencode.CHAR_INT2:
                return read() << 8 | read();
            case Rencode.CHAR_INT4:
                return fetchInt();
            case Rencode.CHAR_INT8:
                return fetchLong();
            default:
                throw new IOException( "Unable to decode the number." );
        }
    }

    private void validate( final int expected )
        throws IOException
    {
        final int token = read();
        if ( token != expected )
        {
            throw new IOException( String.format( "Invalid code: expected=%s, actual=%s", expected, token ) );
        }
    }

    @Override
    public byte readByte()
        throws IOException
    {
        return (byte) readNumber();
    }

    @Override
    public int readUnsignedByte()
        throws IOException
    {
        return (int) readNumber();
    }

    @Override
    public short readShort()
        throws IOException
    {
        return (short) readNumber();
    }

    @Override
    public int readUnsignedShort()
        throws IOException
    {
        return (int) readNumber();
    }

    @Override
    public char readChar()
        throws IOException
    {
        return (char) readShort();
    }

    @Override
    public int readInt()
        throws IOException
    {
        return (int) readNumber();
    }

    @Override
    public long readLong()
        throws IOException
    {
        return readNumber();
    }

    @Override
    public float readFloat()
        throws IOException
    {
        validate( Rencode.CHAR_FLOAT32 );
        return Float.intBitsToFloat( fetchInt() );
    }

    @Override
    public double readDouble()
        throws IOException
    {
        validate( Rencode.CHAR_FLOAT64 );
        return Double.longBitsToDouble( fetchLong() );
    }

    @Override
    public String readLine()
        throws IOException
    {
        return readUTF();
    }

    @Override
    public String readUTF()
        throws IOException
    {
        final int token = read();
        if ( Rencode.STRING_FIXED_START <= token && token < ( Rencode.STRING_FIXED_START + Rencode.STRING_FIXED_COUNT ) )
        {
            return fetchString( token - Rencode.STRING_FIXED_START );
        }
        else if ( '0' <= token && token <= '9' )
        {
            final int size = fetchLength( token );
            return fetchString( size );
        }

        throw new IOException( "Unable to read a String from stream." );
    }

    public List<Object> readList()
        throws IOException
    {
        final int token = read();
        if ( Rencode.LIST_FIXED_START <= token && token < ( Rencode.LIST_FIXED_START + Rencode.LIST_FIXED_COUNT ) )
        {
            int count = token - Rencode.LIST_FIXED_START;
            final List<Object> list = new ArrayList<>( count );
            while ( count-- > 0 )
            {
                list.add( readObject() );
            }

            return list;
        }
        else if ( token == Rencode.CHAR_LIST )
        {
            final List<Object> list = new ArrayList<>( 2 * Rencode.LIST_FIXED_COUNT );
            do
            {
                list.add( readObject() );
            }
            while ( peek() != Rencode.CHAR_TERMINATION );
            read();

            return list;
        }

        throw new IOException( "Unable to read a List from stream." );
    }

    public Map<Object, Object> readMap()
        throws IOException
    {
        final int token = read();
        if ( Rencode.DICTIONARY_FIXED_START <= token && token < ( Rencode.DICTIONARY_FIXED_START + Rencode.DICTIONARY_FIXED_COUNT ) )
        {
            int count = token - Rencode.DICTIONARY_FIXED_START;
            final Map<Object, Object> map = new HashMap<>( count, 1 );
            while ( count-- > 0 )
            {
                map.put( readKey(), readObject() );
            }

            return map;
        }
        else if ( token == Rencode.CHAR_DICT )
        {
            final Map<Object, Object> map = new HashMap<>( 2 * Rencode.DICTIONARY_FIXED_COUNT );
            do
            {
                map.put( readKey(), readObject() );
            }
            while ( peek() != Rencode.CHAR_TERMINATION );
            read();

            return map;
        }

        throw new IOException( "Unable to read a List from stream." );
    }

    private Object readKey()
        throws IOException
    {
        final Object key = readObject();
        if ( key instanceof Number )
        {
            return ( (Number) key ).intValue();
        }
        else if ( key instanceof byte[] )
        {
            return new String( (byte[]) key, Rencode.UTF_8 );
        }

        return key;
    }

    private int peek()
        throws IOException
    {
        final int token = read();
        unread( token );
        return token;
    }

    private int fetchInt()
        throws IOException
    {
        readFully( buffer, 0, 4 );
        return ( buffer[0] & 0xFF ) << 24 | ( buffer[1] & 0xFF ) << 16 | ( buffer[2] & 0xFF ) << 8 | ( buffer[3] & 0xFF );
    }

    private long fetchLong()
        throws IOException
    {
        readFully( buffer, 0, 8 );
        return ( buffer[0] & 0xFFl ) << 56 | ( buffer[1] & 0xFFl ) << 48 | ( buffer[2] & 0xFFl ) << 40 | ( buffer[3] & 0xFFl ) << 32 | ( buffer[4] & 0xFFl ) << 24 | ( buffer[5] & 0xFF ) << 16
            | ( buffer[6] & 0xFF ) << 8 | ( buffer[7] & 0xFF );
    }

    private int fetchLength( int token )
        throws IOException
    {
        final StringBuilder stringBuilder = new StringBuilder( Rencode.MAX_INT_LENGTH );
        stringBuilder.append( (char) token );
        while ( ( token = read() ) != Rencode.LENGTH_DELIMITER )
        {
            stringBuilder.append( (char) token );
        }
        return ( StringUtils.isNumeric( stringBuilder.toString() ) ? Integer.parseInt( stringBuilder.toString() ) : 0 );
    }

    private String fetchString( int size )
        throws IOException
    {
        final byte[] bytes = size <= buffer.length ? buffer : new byte[size];
        readFully( bytes, 0, size );
        return new String( bytes, 0, size, Rencode.UTF_8 );
    }

}
